#include "AnalysisWorkerGroup.h"
#include "AnalysisAlgorithm.h"
#include "AnalysisAlgorithmMap.h"
#include "AnalysisService.h"
#include "DcsInterface.h"

#include "Sct/IS/IOManagerIS.h"
#include "Sct/ISProxy/IOManagerISProxy.h"
#include "Sct/IS/IONameIS.h"
#include "Sct/StdExceptionWrapper.h"
#include "Sct/SctNames.h"
#include "Sct/UniqueID.h"
#include "SctData/TestResult.h"
#include "SctData/DcsData.h"

#include "CalibrationController/IS/TestData.h"
#include <boost/timer.hpp>
#include <iomanip>
#include <is/info.h>
#include <is/infodictionary.h>
#include <is/infoiterator.h>
#include <is/inforeceiver.h>
#include <is/infoT.h>

using namespace SctService;
using namespace boost;
using namespace std;
using namespace Sct;
using namespace Sct::IS;
using namespace SctData;

namespace SctAnalysis {

//TestAlgs methods
AnalysisWorkerGroup::TestAlgs::TestAlgs(boost::shared_ptr<const TestData> testdata) : m_testdata(testdata){
  bool dcsProblemsReported=false;
  for (unsigned imodule=0; imodule<m_testdata->modules_size; ++imodule){
    // make dcs snapshots of all modules in test
    try{
      const std::string& module = m_testdata->modules[imodule];
      boost::shared_ptr<AnalysisAlgorithm> alg = findAlgorithm(module);
      if (!alg.get()) continue; 
      alg->checkTestResultExists();

      // try to get DcsInformation for this module
      try{
	if (!AnalysisService::instance().getDcsInterface().getConfigurationInterface().empty()) {  
	  shared_ptr<SctData::DcsData> dcs (AnalysisService::instance().getDcsInterface().getData(module) );
	  
	  alg->getTestResult()->setDcsData(dcs);
	}
      }catch(Sct::Throwable& e){
	if (!dcsProblemsReported) {
	  e.sendToMrs(MRS_DIAGNOSTIC);
	  dcsProblemsReported=true;
	}
      }catch(std::exception& e){
	if (!dcsProblemsReported) {
	  StdExceptionWrapper(e,__FILE__,__LINE__).sendToMrs(MRS_DIAGNOSTIC);
	  dcsProblemsReported=true;
	}
      }
    }catch(Sct::Throwable& e){
      e.sendToMrs();
    }
  }
}

AnalysisWorkerGroup::TestAlgs::~TestAlgs() {
    boost::recursive_mutex::scoped_lock lock(m_access);

    //We need to explicitly delete the list here so the lock is in force
    //Note that this just removes the pointer..actual deletion will only occur
    //if noone else has grabbed a reference to the object
    m_algorithms.clear();
}

void AnalysisWorkerGroup::TestAlgs::removeAlgorithm(shared_ptr<AnalysisAlgorithm> alg) {
    boost::recursive_mutex::scoped_lock lock(m_access);
    list<shared_ptr<AnalysisAlgorithm> >::iterator it =
        find(m_algorithms.begin(), m_algorithms.end(), alg);

    if ( it != m_algorithms.end() )
        m_algorithms.erase(it);
}

shared_ptr<const TestData>AnalysisWorkerGroup::TestAlgs::getTest() const{
  if (!m_testdata.get()) throw Sct::InvariantViolatedError("Attempt to retrieve null TestData - object is not properly constructed",__FILE__,__LINE__);
  return m_testdata;
}

ostream& AnalysisWorkerGroup::TestAlgs::printStatus(ostream& os) const throw() {
    boost::recursive_mutex::scoped_lock lock(m_access);
    os << setw(26) << getTest()->testName 
       << setw(8) << getTest()->runNumber
       << setw(8) << getTest()->startScanNumber
       << setw(8) << getTest()->nScans 
       << setw(26) << getTest()->analysisAlgorithm
       <<  endl;

    return os;
}

shared_ptr<AnalysisAlgorithm> AnalysisWorkerGroup::TestAlgs::findAlgorithm(const string& modulename) {
    boost::recursive_mutex::scoped_lock lock(m_access);
    shared_ptr<AnalysisAlgorithm> alg;
    for (list<shared_ptr<AnalysisAlgorithm> >::const_iterator it=m_algorithms.begin();
            it != m_algorithms.end();
            ++it) {
        if ( (*it)->getModuleName()==modulename ) {
            alg=(*it);
            break;
        }
    }
    if (!alg) {
        alg=addAlgorithm(modulename);
    }
    return alg;
}

shared_ptr<AnalysisAlgorithm> AnalysisWorkerGroup::TestAlgs::addAlgorithm(const string& modulename) {
    boost::recursive_mutex::scoped_lock lock(m_access);
    shared_ptr<AnalysisAlgorithm> alg = AnalysisAlgorithmMap::instance().getAlgorithm(getTest(), modulename);
    m_algorithms.push_back(alg);
    return alg;
}

//AnalysisWorkerGroup methods


ostream&  AnalysisWorkerGroup::printStatus(ostream& os) const throw() {
    boost::recursive_mutex::scoped_lock lock(m_tests_access);
    os << "\t === Analysis Service ===\n" << endl;
    os << "AnalysisWorkerGroup tests:" << endl;
    unsigned ntest=0;
    os << setw(26) << "TestName" 
       << setw(8) << "Run"
       << setw(8) << "Scan"
       << setw(8) << "NScans" 
       << setw(26) << "AnalysisAlg"
       <<  endl;
    os << "----------------------------------------------------------------"<<endl;
    for (list<shared_ptr<TestAlgs> >::const_iterator it = m_tests.begin();
            it != m_tests.end() ; ++it ) {
        (*it)->printStatus(os);
        ++ntest;
    }
    os << "----------------------------------------------------------------"<<endl;
    os << endl;
    os << "\tWorkers=" << nWorkers() << "\t  (Busy=" << busy() << ")"
       << "\tQueue =" << queueSize();
    if (isFifo()) {
      os << " [FIFO]\n";
    }else{
      os << " [FILO]\n";
    }
    os <<"Total number of tests = " << ntest << endl;
    return os;
}

void AnalysisWorkerGroup::addTest(shared_ptr<const TestData> testdata) {
    boost::recursive_mutex::scoped_lock lock(m_tests_access);
    m_tests.push_back(shared_ptr<TestAlgs>(new TestAlgs(testdata) ) );
}

void AnalysisWorkerGroup::removeTestsUpTo(shared_ptr<const TestData> testdata) {
    boost::recursive_mutex::scoped_lock lock(m_tests_access);
    list<shared_ptr<TestAlgs> >::iterator it = m_tests.begin();

    while( it != m_tests.end() ) { // no ++it here since we need to delete it in the function!
        if ((*it)->getTest()->runNumber==testdata->runNumber
                && (*it)->getTest()->startScanNumber<=testdata->startScanNumber) {
            
	  cout << "Deleting old test for run "
	       << (*it)->getTest()->runNumber << ", start scan "
	       << (*it)->getTest()->startScanNumber
	       << endl;
            it=m_tests.erase(it);   // erase returns ++it
        } else {
            ++it;                   // increment ourselves
        }
    }
}

shared_ptr<AnalysisWorkerGroup::TestAlgs> AnalysisWorkerGroup::findTest(const TestData& testdata) const throw() {
    return findTest(testdata.runNumber, testdata.startScanNumber);
}

shared_ptr<AnalysisWorkerGroup::TestAlgs> AnalysisWorkerGroup::findTest(const unsigned long runno, const unsigned long scanno) const throw() {
    boost::recursive_mutex::scoped_lock lock(m_tests_access);
    shared_ptr<AnalysisWorkerGroup::TestAlgs> talgs;
    for (list<shared_ptr<TestAlgs> >::const_iterator it = m_tests.begin();
            it != m_tests.end();
            ++it ) {
        shared_ptr<const TestData> td=(*it)->getTest();
        if ( runno == td->runNumber &&
                scanno >= td->startScanNumber &&
                scanno < td->startScanNumber + td->nScans) {
            talgs=*it;
            //		cout << "found test for run " << runno<< ", scan " << scanno << endl;
        }
    }
    return talgs;
}




void AnalysisWorkerGroup::purge() throw() {
    //Stop the workers from trying to get on with things
    this->setPaused(true);
    while (paused() != nWorkers()) sleep(1);
    
    // Clear the queue.
    shared_ptr<IOName> s;
    do {s=pop(); } while(s);
    boost::recursive_mutex::scoped_lock lock(m_tests_access);

    // Clear the test data.
    cout << "Clearing tests ... " << endl;
    m_tests.clear();
    cout << " ... done clearing tests" << endl;
    this->setPaused(false);    
}

void AnalysisWorkerGroup::work(shared_ptr<IOName> name) throw() {
    using namespace SctData;

    try {
    /// (1) special case for control data callbacks
      if (name->getIOName().find("TestData") != string::npos){
	boost::shared_ptr<TestData> testdata( new TestData() );
	ISInfoDictionary& id = Sct::SctNames::getISDictionary();
	ISInfo::Status result = id.findValue(name->getIOName().c_str(), *testdata);
	if (result != ISInfo::Success) {
	  throw IsException(result,std::string("Error reading from IS server.  Couldn't read: ")
			    + name->getIOName(), __FILE__, __LINE__);	
	}
	if (findTest(*testdata).get() == 0 ) {
	  addTest(testdata);
	} else {
	  throw Sct::IoException(std::string("AnalysisService::TestCallback Test already exists: ")
				 +name->getIOName(), __FILE__, __LINE__);
	}
	return;
      }
	
      // (2) Other case is callback for scan data 
	timer t;
	UniqueID id (name->getUniqueID());
        const unsigned runno = id.getRunNumber();
        const unsigned scanno = id.getScanNumber();
        const string modulename= id.getModule();

        //	    cout << "run="<< runno << ", scan="<< scanno << ", modulename="<<modulename
        //		 << ", classname="<<isname.getClassName()<<endl;

        shared_ptr<AnalysisWorkerGroup::TestAlgs> talgs = findTest(runno, scanno);

        //	    cout << "Found test" << endl;

        if (!talgs) {
           std::ostringstream oss; oss << "AnalysisWorker::work. No test found for run " 
		                       << runno << " scan" << scanno;
    		throw IllegalStateError(oss.str(), __FILE__, __LINE__);
        }
        // Get a testresult. Add a new testresult if one dosen't exist for this module
        shared_ptr<AnalysisAlgorithm> algorithm = talgs->findAlgorithm(modulename);
	if (!algorithm.get()) return;

        //	    cout << "algorithm.get()=" << algorithm.get() << endl;

        bool isfit=(name->getClassName()=="SctData::FitScanResult");

        // lock the approprate result and do your stuff, analysis-algorithm.
        boost::recursive_mutex::scoped_lock lock (algorithm->getMutex());
        if (isfit) {
            //		cout << "got fit from queue" << endl;
            algorithm->addFitScanResult(name);
        } else {
            //		cout << "got raw from queue" << endl;
            algorithm->addRawScanResult(name);
        }
	if (algorithm->canAnalyze()) {
	    if (algorithm->isDone()) throw IllegalStateError("AnalysisWorker::work.  Algorithm has additional, unnecessary Scan - analysis already done", __FILE__, __LINE__);
	    algorithm->checkTestResultExists();
	    algorithm->addOverheadTime(t.elapsed());
	    
	    t.restart();
	    algorithm->loadData();
	    algorithm->addIOTime(t.elapsed());
	    
	    t.restart();
	    algorithm->analyze();
	    algorithm->addAnalysisTime(t.elapsed());
	    
	    t.restart();
	    algorithm->finish();
	    algorithm->addIOTime(t.elapsed());
	    talgs->removeAlgorithm(algorithm);
	} else {
	    algorithm->addOverheadTime(t.elapsed());
	}
	
    } catch (InvalidArgumentError& e) {
	e.sendToMrs();
    } catch(Throwable& e) {
        e.sendToMrs(MRS_ERROR);
    } catch(std::exception& e) {
	StdExceptionWrapper sew(e);
        sew.sendToMrs(MRS_ERROR);        
    } catch(...) {
	Error e("uncaught unknown exception", __FILE__, __LINE__);
	e.sendToMrs(MRS_ERROR);
    }
}


}// end of namespace SctAnalysis
