#include "StrobeDelayAlgorithm.h"
#include "AnalysisAlgorithmMap.h"
#include "SctData/StrobeDelayTestResult.h"
#include "SctData/FitScanResult.h"
#include "SctData/FitObject.h"
#include "SctData/TopHatFitObject.h"
#include "SctData/DefectList.h"
#include "SctData/StandardDefects.h"
#include <string>

using namespace std;
using namespace boost;
using namespace SctData;
using namespace Sct;

namespace SctAnalysis {
    
bool StrobeDelayAlgorithm::inMap = AnalysisAlgorithmMap::instance().setAlgorithm("StrobeDelayTest", auto_ptr<AnalysisAlgorithm>(new StrobeDelayAlgorithm()));

shared_ptr<AnalysisAlgorithm> StrobeDelayAlgorithm::clone(shared_ptr<const TestData> testData, const string& moduleName) const throw(){
    return shared_ptr<AnalysisAlgorithm>(new StrobeDelayAlgorithm(testData, moduleName, *this));
}

shared_ptr<TestResult> StrobeDelayAlgorithm::createTestResult() const {
    return shared_ptr<StrobeDelayTestResult> (new StrobeDelayTestResult()) ;
}

void StrobeDelayAlgorithm::loadData() {
    loadAllFits();
}

bool StrobeDelayAlgorithm::canAnalyze() const {
    return hasAllFits();
}

void StrobeDelayAlgorithm::analyze() {
    const double fraction=0.4;
    StrobeDelayTestResult& result = *dynamic_pointer_cast<StrobeDelayTestResult>(getTestResult());
    const FitScanResult& fitted = *getFit(0);
    result.setScanVariable(fitted.getHeader().getVariable());	

    bool pass = true;
    bool problem = false;
    result.setNOptima(nChipModule);
    for (short unsigned int ichip=0; ichip<nChipModule; ++ichip ) {
      const DefectSeverity& severity = result.getDefects().defectSeverityAffectingElement(ModuleElement::Chip(ichip));
      if (severity>=SERIOUS) pass=false;
      FitObject& fit = fitted.getChipFit(ichip);
      try {
	dynamic_cast<TopHatFitObject&> (fit);
      } catch(std::exception& e) {
	throw InvariantViolatedError("StrobeDelayAlgorithm::optimize ERROR not a TopHat function", __FILE__, __LINE__);
      }

      double rise_value=fit.getParameter(1), fall_value=fit.getParameter(3);
      double rise_width=fit.getParameter(2), fall_width=fit.getParameter(4);

      result.setOptimum(ichip, rise_value + fraction*(fall_value-rise_value));

      if (rise_value < 0) {
	result.getDefects().addDefect(Defect(StandardDefects::SD_LO, ModuleElement::Chip(ichip)));
	pass = false;
	ostringstream msg;
	msg << "Chip (0-11) #" << ichip << " mid-rise is at " << rise_value;
	result.addComment(msg.str());
      } else if (rise_value-rise_width < 1){
	problem=true;
	ostringstream msg;
	msg << "Chip (0-11) #" << ichip << " 1-sigma before mid-rise is at " << rise_value-rise_width;
	result.addComment(msg.str());
      }
      if (fall_value > 63) {
	result.getDefects().addDefect(Defect(StandardDefects::SD_HI, ModuleElement::Chip(ichip)));
	pass = false;
	ostringstream msg;
	msg << "Chip (0-11) #" << ichip << " mid-fall is at " << fall_value;
	result.addComment(msg.str());
      } else if (fall_value+fall_width>62){
	problem=true;
	ostringstream msg;
	msg << "Chip (0-11) #" << ichip << " 1-sigma after mid-fall is at " << fall_value+fall_width;
	result.addComment(msg.str());
      }
    }
    result.setPassed(result.getDefects().nChannelsAffected()<(128*2));
    result.setProblem(problem);
}


}
