#ifndef SCTDATA_ISSUMMARYMANAGER_H
#define SCTDATA_ISSUMMARYMANAGER_H
#include <boost/utility.hpp>
#include <map>
#include <string>
#include <boost/shared_ptr.hpp>

using std::map;

namespace SctData {
    class TestResult;
    /**
       Classes which convert SctData::TestResult data into SctDaq style summary files.
       The output is to a output stream, which is given to the ISSummaryManager. This output stream could be e.g. std::cout, or an output string stream, or perhaps the output file stream created by ISSummaryManager::getOutputFileStream(moduleName);
       Example code snippet:
       <PRE>
       ostringstream datastream;
       shared_ptr<const SctData::TestISSummary::ISSummaryWriter> writer =
       SctData::TestISSummary::ISSummaryManager::instance().getWriter(myTestResult.getClassName());
       writer->writeHeader(*m_result, datastream);
       writer->write(*m_result, datastream);
       myString = datastream.str();
       </PRE>
       
    */
    namespace ISTestSummary{
      class ISSummaryWriter;
      /**
	   Singleton manager provides the public interface to writing a summary for 
	   a particular TestResult
	*/

	class ISSummaryManager : boost::noncopyable {
	public:
	    /** get the instance of the singleton*/ 
	  static ISSummaryManager& instance();
	    
	    /**
	      For use by clients
	      Writes to a given ostream the summary.
	    */
	  void write(const TestResult& t) ;
	  
	  /** 
	      concrete ISSummaryWriters should add themselves to the map using this mechanism 
	      @param testname the name of the TestResult for which this writer will write
	      @param writer a shared pointer to the concrete ISSummaryWriter 
	  */
	  bool addWriter(const std::string& testname, boost::shared_ptr<ISSummaryWriter> writer);
	  /** 
	      get the appropriateISSummaryWriter for the TestResult
	      @param classname the name of the TestResult for which the summary is to be written.
	  */
	  boost::shared_ptr<const ISSummaryWriter> getWriter(const std::string& classname) const;
	private:
      	    ~ISSummaryManager(); ///< do NOT delete the singleton! 
	    ISSummaryManager(); ///< Singleton.
	    
	    typedef map<std::string, boost::shared_ptr<const ISSummaryWriter> > WriterMap;
	    /** 
		maps TestResult names to ISSummaryWriters 
	    */
	    WriterMap theMap;
	};
    }
} // end of namespace 

#endif // #ifndef SCTDATA_ISSUMMARYMANAGER
