/************************************************************************************
 * slaveTasks.c
 *   
 *  synopsis: This file contains tasks which run on the slave DSPs. Typically these
 *            tasks process event data as it arrives.
 *
 *  in this file: histogramTask, trapTask, errorTask, occupancyTask, resynchTask 
 *
 *  related files:
 *   taskManager.c:  Contains routines for starting, stopping & managing tasks.
 *   masterTasks.c:  Contains the master DSP tasks.
 *   histogram.h:    Contains the definitions for the histogram control structures
 *                   on the master & slave DSPs.
 ************************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <csl_cache.h>

#include "resources.h"
#include "memoryPartitions.h"
#include "taskManager.h"
#include "rodConfiguration.h"
#include "comRegDfns.h"
#include "histogram.h"
#include "registerIndices.h"

extern char genStr[];
extern TIMER_Handle timer1;  /* general purpose timer, started in main() */
extern far HistoCtrl histoCtrl;
extern far BinCtr    binCtr;
extern far ErrorInfo *errorInfo;

#pragma CODE_SECTION(histogramTask,   "icode");
#pragma CODE_SECTION(trapTask,        "icode");
#pragma CODE_SECTION(occupancyTask,   "xcode");
#pragma CODE_SECTION(errorTask,       "xcode");
#pragma CODE_SECTION(resynchTask,     "xcode");

/* Histogram Task state-machine states: */
/* TASK_INIT == 1 */
#define  HISTO_READY        0x2
#define  HISTO_EXPECTING    0x3
#define  HISTO_LOCAL_INC    0x4
#define  HISTO_LOCAL_TRIG   0x5

#include <csl.h>
#include <csl_mcbsp.h>
typedef struct {
	UINT32 code, loopCnt, timeStamp, procTime;
	UINT32 trapStat[3], isrState;
	UINT32 nEvents, nHistoEvents, p4MaskStat, maskFlag;
} GenEventInfo;

GenEventInfo *genEventInfo= (GenEventInfo *) 0xb0000100;

typedef struct { //TNV
	UINT32 *badcafe_address;
} GenEventInfoTail;

GenEventInfoTail *genEventInfoTail = (GenEventInfoTail *) 0xb0000000; //TNV

/************************************************************************************
 * histogramTask
 *   
 *   synopsis: histogramming task. 
 * 
 *   author: Douglas Ferguson
	#define R_HISTOGRAM_TASK  100
	struct HISTOGRAM_TASK_IN {
		UINT32 nEvents, controlFlag;  * nEvents can equal COLLECT_FOREVER too. 
	};

	#define MASTER_HREG     0
	#define LOCAL_INC       1
	#define LOCAL_SET_TRIG  2

	struct HISTOGRAM_TASK_OUT {
		UINT32  nEvents, binsDone;
		FLOAT32 avgProcTime;
		UINT32  *dataPtr, dataLength;
	};
 ************************************************************************************/
INT32 histogramTask(TaskInput *tsi, TaskOutput *tso,
                    UINT32 opFlags) {
	static INT32    returnCode, errorCode;
	static UINT8    state= TASK_INIT, controlFlag, first, set;
	static UINT32   nEvents, nEventsBin, nTotBin, nErrorsBin,
	                nErrorMask, idleCnt;
	static FLOAT32  avgProcTime;
	static UINT16   binsDone, bin;
	struct   EventData *data;
	UINT32   procTime, t0= 0, t1= 0, t3= 0;
	UINT32   val, fmt;

	/********************* data retrieval & routine reset *********************/
	if (opFlags > 0x10) {
		/* Place the interesting internal variables into the output structure */
		tso->histogramTaskOut.nEvents=     nEvents;
		tso->histogramTaskOut.binsDone=    binsDone;
		tso->histogramTaskOut.avgProcTime= avgProcTime;
		tso->histogramTaskOut.dataPtr=     histoCtrl.base;
		tso->histogramTaskOut.dataLength=  histoCtrl.size;

		if (opFlags & TASK_HALT_MASK) {
			newInformation(__FILE__, __LINE__, "Histogram task halting.\n");
			// Registers used for mdsp/sdsp flow control may be in random states:
			// here we reset them to the correct defaults. PWP 10/11/05
			RST_RBIT(HSTAT_REG_0, HSR0_SLV_RDY);  // not ready
			RST_RBIT(HSTAT_REG_0, HSR0_SLV_PROC); // not processing
			RST_RBIT(HSTAT_REG_0, HSR0_SLV_EXP);  // not expecting
			RST_RBIT(HSTAT_REG_0, HSR0_SLV_DONE); // not done
			state= TASK_INIT;
			return TASK_HALTED;
		}
		else if (opFlags & TASK_INIT_MASK) {
			newInformation(__FILE__, __LINE__, "Histogram task re-initing.\n");
			state= TASK_INIT;
			return TASK_INITIALIZED;
		}
		else if (opFlags & TASK_QUERY_MASK) {
			newInformation(__FILE__, __LINE__, "Histogram task query data output.\n");
			return TASK_QUERIED;
		}
	}

	/*************************** routine states:INIT **************************/
	if (state == TASK_INIT) {
		/* set up any static variables beyond state */
		if (GET_RBIT(DIAGNOSTIC_REG, DR_INFO_DISPLAY)) {
			newInformation(__FILE__, __LINE__, "state -> INIT \n\n.");
		}
		returnCode= REPEAT_TASK;
		nEvents= 0;    binsDone= 0;    first= TRUE;
		avgProcTime= 0.00;         nEventsBin= nErrorsBin= nTotBin= 0;
		controlFlag= tsi->histogramTaskIn.controlFlag;

		if (!histoCtrl.histogramsAreSetup) { /* return error */
			sprintf(genStr, "%s%s","Histograms have not been defined yet. ",
				"Run histogramSetup primitive and re-start.\n\n");
			newError(&returnCode, NO_HISTOGRAMS_ERROR, FATAL_ERR, "histogramTask",
			         genStr, __FILE__, __LINE__);
			return returnCode;
		}
		set= histoCtrl.histogramSet;
		bin= histoCtrl.currentBin;

		tso->histogramTaskOut.nEvents=     nEvents;
		tso->histogramTaskOut.binsDone=    binsDone;
		tso->histogramTaskOut.avgProcTime= avgProcTime;
		tso->histogramTaskOut.dataPtr=     histoCtrl.base;
		tso->histogramTaskOut.dataLength=  histoCtrl.size;

		//PWP 10/11/05
		RST_RBIT(HSTAT_REG_0, HSR0_SLV_PROC); // not processing
		RST_RBIT(HSTAT_REG_0, HSR0_SLV_EXP);  // not expecting
		RST_RBIT(HSTAT_REG_0, HSR0_SLV_DONE); // not done


		nErrorMask= READ_REG(RESERVED_REG_0);

		if (controlFlag == MASTER_HREG) {
			SET_RBIT(HSTAT_REG_0, HSR0_SLV_RDY);
			state= HISTO_READY;
		}
		else if (controlFlag == LOCAL_INC) {
			state= HISTO_LOCAL_INC;
			/* ASGN_RFLD(HSTAT_REG_0,HSR0_SLV_SET,HSR0_SLV_SET_W,set); */			
			ASGN_RFLD(HCMD_STAT_REG_0, HCSR0_SLV_BIN, HCSR0_SLV_BIN_W, bin);			
		}
		else if (controlFlag == LOCAL_SET_TRIG) {
			state= HISTO_LOCAL_TRIG;
		}
		setTaskState(HISTOGRAM_TASK, state);
	}

	/************************* routine states:LOCAL_TRIG **********************/
	if (state == HISTO_LOCAL_TRIG) {
		if (first) {
			first= FALSE;
			newInformation(__FILE__, __LINE__, "state -> LOCAL_TRIG \n\n.");
			setTaskState(HISTOGRAM_TASK, state);
		}

		/* bin & set are set by a locally running setTrigger, need only update
		   the variables. */
		set= histoCtrl.histogramSet;
		bin= histoCtrl.currentBin;
		/* ASGN_RFLD(HSTAT_REG_0,HSR0_SLV_SET,HSR0_SLV_SET_W,set); */			
		ASGN_RFLD(HCMD_STAT_REG_0, HCSR0_SLV_BIN, HCSR0_SLV_BIN_W, bin);			
	}

	/***************************** Event Processing ***************************/
	data= getNextEvent();
	if (data != NULL) {
		//If an event is ready, update the idle counter in the status register:
		WRITE_REG(HSTAT_REG_2, idleCnt);   idleCnt= 0;
		
		if (state == HISTO_READY) {
			if (!(data->flags & OBSERVATION_STRAY_MASK)) {
				data->flags |= OBSERVATION_STRAY_MASK;
				sprintf(genStr, "%s%s%s%s%d%s", "An unexpected event arrived without ",
				                "the Master DSP indicating that it was sending ",
				                "triggers. Event  marked as a stray event. Internal ",
				                "ID= ",data->id,".\n");
				newError(&errorCode, UNEXPECTED_EVENT, ERROR_0, "histogramTask",
				         genStr, __FILE__, __LINE__);
			}

			eventInfo0M;
		}

		/* if errors, print warning message and leave event as a marked stray; the
		   event manager will then copy it to the reserve buffer for observation. */
		else if (data->error > 0) {

			if (!(data->flags & OBSERVATION_STRAY_MASK)) {
//				data->flags |= OBSERVATION_STRAY_MASK;

				val = READ_REG(EFB_RUNTIME_STAT_REG);
				sprintf(genStr,"EFB_RUNTIME_STAT_REG %08x FMT_HR_ERR",val);
				
				for(fmt=0;fmt<4;fmt++){
					val = READ_REG(FMT_HEADER_TRAILER_ERR(fmt));
					sprintf(genStr,"%s (%d) %08x",genStr,fmt, val);
				}
				newInformation(__FILE__,__LINE__,genStr);
									
				fsrp0_on;
					SET_RBIT(HSTAT_REG_0, HSR0_SLV_PROC);
				fsrp0_off;
								
				sprintf(genStr,"EVENT_ERROR bin %d n_good_events %d n_bad_events %d",
				bin, nEventsBin, nErrorsBin);
				newInformation(__FILE__,__LINE__,genStr);

				++nErrorsBin;
				ASGN_RFLD(HSTAT_REG_0, HSR0_SLV_BINERR, HSR0_SLV_BINERR_W, nErrorsBin);			
				eventInfo1M;

				histoError_c(data);

				fsrp0_on;			
					RST_RBIT(HSTAT_REG_0, HSR0_SLV_PROC);
				fsrp0_off;

				// Flush cache to make updated errorInfo available to mdsp
				CACHE_flush(CACHE_L2ALL, (void*) 0x0, 0x0);	
			}
//#else
#ifdef FLYING_BLIND

			if (!(data->flags & OBSERVATION_STRAY_MASK)) {
				data->flags |= OBSERVATION_STRAY_MASK;
				//if (data->nErrors != nErrorMask) {
					sprintf(genStr, "%s",
					  "Event contains errors; marked as a stray event.\n");
					newError(&errorCode, HISTOGRAMMING_ERROR_EVT, ERROR_0,
					         "histogramTask", genStr, __FILE__, __LINE__);
					         
					sprintf(genStr, "\n %s0x%08x%s %s0x%08x%s %s0x%08x%s ", 
		               "   Current TS0= ", READ_REG(TRAPSTAT_REG1_0), ".\n",
		               "   Current TS1= ", READ_REG(TRAPSTAT_REG1_1), ".\n",
		               "   Current TS2= ", READ_REG(TRAPSTAT_REG1_2), ".\n");

					newInformation(__FILE__,__LINE__,genStr); // PWP 01.11.05
			//}
				++nErrorsBin;
				ASGN_RFLD(HSTAT_REG_0, HSR0_SLV_BINERR, HSR0_SLV_BINERR_W, nErrorsBin);			
			}

			eventInfo1M;

#endif
		}

		else {   /* for all other states, process the event */
			#if 0
			/* informational histogramming fxn */
			//if (GET_RBIT(DIAGNOSTIC_REG, DR_DISP_LINK_DATA)) histoEvent_list(data);

			/* non- optimized histogramming fxn */
			t1= 0;
			if (GET_RBIT(DIAGNOSTIC_REG, DR_HISTO_1)) {
			fsrp0_on;			
				t0= TIMER_getCount(timer1);
				//histoEvent1(data);
				t1= (delta_t(t0))/10;
			fsrp0_off;			
			}

			delay(1);
			#endif

			fsrp0_on;			
				++nEventsBin;  ++nEvents,
				++binCtr.ctr[bin];
				SET_RBIT(HSTAT_REG_0, HSR0_SLV_PROC);
			fsrp0_off;			

			t0= TIMER_getCount(timer1);
			fsrp0_on;			

			eventInfo2M;
			histoCtrl.histoFxn(data);

			fsrp0_off;	fsxp0_toggle;		
			procTime= delta_t(t0);
			fsxp0_toggle;		
			eventInfo3M;

			delay(1);

			fsrp0_on;			
				WRITE_REG(HSTAT_REG_1, nEventsBin);
				WRITE_REG(HSTAT_REG_3, procTime);			
				RST_RBIT(HSTAT_REG_0, HSR0_SLV_PROC);
			fsrp0_off;			


			if (GET_RBIT(DIAGNOSTIC_REG, DR_INFO_DISPLAY)) {
				sprintf(genStr, "%s%d%s %s%d%s %s0x%04x%s %s%d, %d and %d%s",
				                "Bin ", bin, ",  ",
				                "event ", nEventsBin, ":   ",
				                "length (w) ", data->length, ",   ",
				                "histo 1  2  3  dt=  ", t1, procTime, t3,
				                " micro-seconds.\n");                
				newInformation(__FILE__, __LINE__, genStr);
			}
		}

		fsrp0_on;			
			/* remove the event from the queue */
			//processEvent(data, Q_HISTOGRAM);
		fsrp0_off;			

	}
	
	//If no event, update the idle counter (written to status reg. upon event):
	else {++idleCnt; }

	/*************************** routine states:READY *************************/
	if (state == HISTO_READY) {
		if (first) {
			first= FALSE;
			if (GET_RBIT(DIAGNOSTIC_REG, DR_INFO_DISPLAY)) {
				newInformation(__FILE__, __LINE__, "state -> READY \n\n.");
			}
			setTaskState(HISTOGRAM_TASK, state);
			nEventsBin= nErrorsBin= 0;
			/* WRITE_REG(HSTAT_REG_1,    0); */
		}

		/* check HCMD_STAT_REG_0. If the new-bin bit is set, prepare for receiving
		   events (state -> EXPECTING). */
		if (GET_RBIT(HCMD_STAT_REG_0, HCSR0_SLV_NEWBIN)) {
			bin= GET_RFLD(HCMD_STAT_REG_0, HCSR0_SLV_BIN, HCSR0_SLV_BIN_W);
			nTotBin= READ_REG(HCMD_STAT_REG_1);

			/* set= GET_RFLD(HCMD_STAT_REG_0, HCMD_SLV_SET, HCMD_SLV_SET_W); */
			set= 0;
			histoCtrl.currentBin= bin;   histoCtrl.histogramSet= set;

			histoCtrl.calLineEnable= GET_RBIT(HCMD_STAT_REG_0, HCSR0_SLV_CALLINE_EN);
			if (histoCtrl.calLineEnable) {
				histoCtrl.calLine= GET_RFLD(HCMD_STAT_REG_0,
				                         HCSR0_SLV_CALLINE, HCSR0_SLV_CALLINE_W);
			}

			setHistoModuleBase(FALSE /* init */, bin);
			WRITE_REG(HSTAT_REG_1, nEventsBin);

			RST_RBIT(HSTAT_REG_0, HSR0_SLV_DONE);
			SET_RBIT(HSTAT_REG_0, HSR0_SLV_EXP);

			first= TRUE;
			state= HISTO_EXPECTING;
		}
	}

	/************************* routine states:EXPECTING ***********************/
	else if (state == HISTO_EXPECTING) {
		fsrp0_on;			

		if (first) {
			if (GET_RBIT(DIAGNOSTIC_REG, DR_INFO_DISPLAY)) {
				newInformation(__FILE__, __LINE__, "state -> EXPECTING \n\n.");
			}
			first= FALSE;
			setTaskState(HISTOGRAM_TASK, state);
		}

		/* check to see if we have gotten all the expected events for this bin. */
		if (nEventsBin >= nTotBin) {
			RST_RBIT(HSTAT_REG_0, HSR0_SLV_EXP);
			SET_RBIT(HSTAT_REG_0, HSR0_SLV_DONE);
			/* WRITE_REG(HCMD_STAT_REG_1, 0); */		

			if (GET_RBIT(DIAGNOSTIC_REG, DR_INFO_DISPLAY)) {
				sprintf(genStr, "%s%d%s%d%s%d%s", "Bin #", bin, ":  nEvents, nErrors= ",
				                              nEventsBin, ", ",nErrorsBin, ".\n");
				newInformation(__FILE__, __LINE__, genStr);
			}

			first= TRUE;
			state= HISTO_READY;
		}

		fsrp0_off;			
	}

	/************************* routine states:LOCAL_INC ***********************/
	else if (state == HISTO_LOCAL_INC) {
		if (first) {
			first= FALSE;
			newInformation(__FILE__, __LINE__, "state -> LOCAL_INC \n\n.");
			setTaskState(HISTOGRAM_TASK, state);
		}

		if (data != NULL) {
			++bin;  bin%= histoCtrl.nBins;
			histoCtrl.currentBin= bin;
			setHistoModuleBase(FALSE /* init */, bin);
			ASGN_RFLD(HCMD_STAT_REG_0, HCSR0_SLV_BIN, HCSR0_SLV_BIN_W, bin);			
		}
	}

	/**************************************************************************/
	if (tsi->histogramTaskIn.nEvents != COLLECT_FOREVER) {
		if (nEvents >= tsi->histogramTaskIn.nEvents) {
			state= TASK_DONE;
			setTaskState(HISTOGRAM_TASK, state);
			return SUCCESS;     /* task removal will gather output information */
		}
	}

	fsrp0_on;   delay(5); fsrp0_off;
	return returnCode;
}


#pragma DATA_SECTION(trapTempOut, "idata");
struct tTO {
	UINT32 marker;
	UINT32 tFlag;
	UINT32 nTrapped;
	UINT32 nTot;
} trapTempOut;

#define  TRAP_REQUEST     0x2
#define  TRAP_EVENTS      0x3
/************************************************************************************
 * trapTask
 *   
 *   synopsis: Traps incoming events (places them in the appropriate spot in memory),
 *             and exits when the requested number of events have been collected.
 *     It is the responsibility of the user to make sure that the events are
 *     transferred (sp?) out of memory in a timely manner, to prevent the trapping
 *     buffer from over-filling. The (nyi) master event trap monitoring routine helps
 *     to coordinate this effort; for now everything is manual.
 *
 *   input parameters:
 *
 *     nEvents:   The number of events to trap. If the special COLLECT_FOREVER value
 *                is input (the same as in eventTrapSetup), events will be sought
 *       until the task is ended. If HOST_TRIGGERED_EVENTS is input, the master DSP
 *       will watch the trap command register for requests from the VME host. (This
 *       is functionally equivalent to, but more convenient and efficient than,
 *       sending a series of 1 event trapping tasks using startTask; it also frees the
 *       master DSPs primitive list for other priorities).
 *
 *     reloadInterval: The interval in task iterations before a trap request is sent
 *                     to the master event trap monitoring routine, once an event has
 *        been trapped. If zero, the trap is put in continuous mode until at least the
 *        number of requested events have been gathered; otherwise the trap is set for
 *        single-shot collecting and the trap reset according to the reload interval.
 *
 *     trapType:  The type of events to be captured: ATLAS, TIM, ROD or ERROR events.
 *                (The same parameters as used for eventTrapSetup). The trapping
 *       task is meant to be run as an auxiliary task, started when necessary & then
 *       running alongside other pre-existing tasks. Therefore the preferred way of
 *       setting the parameters in eventTrapSetup is to leave the 2nd trap idle, and
 *       input the needed parameters here. If both traps have already been set by the
 *       ets primitive, the trap configuration input here must be idle, and at least
 *       one of the two traps must be pre-configured to send events to the trapping
 *       task. 
 *
 *     eventType: The type of ATLAS, TIM or ROD events desired.
 *
 *     recycle (nyi):  If set, the task recycles events after it has filled its buffer;
 *                     meant to be used in conjunction with the master monitoring
 *       routine, which will relay data locations to the host.
 *
 *     trapAttempts (pi):  The number of attempts the task will make to trap the
 *                          event before discarding it. This can be shorter than the
 *       event manager iteration limit; if it is longer the event will be treated as
 *       a stray event.
 *
 *     trapBufferBase, trapBufferLength: The location & length of memory to use as a
 *                                       trapping buffer; defaults are supplied. It is
 *       the user's responsibility to avoid over-writing other critical data; later
 *       when a more useful form of memory allocation is done this can be tested for.   
 *
 *     override (nyi): For later use.
 * 
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 trapTask(TaskInput *tsi, TaskOutput *tso,
               UINT32 opFlags) {
	static INT32  returnCode;
	static UINT8  state= TASK_INIT, first;
	static UINT32 compLen;
	static UINT32 *bufferBase, bufferLen, nEvents, eventType, eventAnalysis;
	/*iteration= 0, lastRequestIter= 0, trapAttempts= 0; */
	static UINT32 dataLen, nEventsTrapped, eventCount;
	struct EventData *data;
	INT32  errorCode;
	UINT32 nErrors, routerError, *dataPtr1, *dataPtr2, i;
	UINT8  trapEvent;

	/********************* data retrieval & routine reset *********************/
	if (opFlags > 0x10) {
		/* Place the interesting internal variables into the output structure */
		tso->trapTaskOut.nEvents=     nEventsTrapped;
		tso->trapTaskOut.dataLen=     dataLen;
		tso->trapTaskOut.bufferBase=  bufferBase;
		tso->trapTaskOut.bufferLen=   bufferLen;

		if (opFlags & TASK_HALT_MASK) {
			newInformation(__FILE__, __LINE__, "Trap task halting.\n");
			state= TASK_INIT;
			return TASK_HALTED;
		}
		else if (opFlags & TASK_INIT_MASK) {
			newInformation(__FILE__, __LINE__, "Trap task re-initing.\n");
			state= TASK_INIT;
			return TASK_INITIALIZED;
		}
		else if (opFlags & TASK_QUERY_MASK) {
			newInformation(__FILE__, __LINE__, "Trap task query data output.\n");
			return TASK_QUERIED;
		}
	}

	/*************************** routine states:INIT **************************/
	if (state == TASK_INIT) {
		/* set up any static variables beyond state */
		newInformation(__FILE__, __LINE__, "state -> INIT \n\n.");
		returnCode= REPEAT_TASK;

		first= TRUE;
		dataLen= nEventsTrapped= eventCount= 0;

		nEvents=    tsi->trapTaskIn.nEvents;
		eventType=  tsi->trapTaskIn.eventType;
		bufferBase= tsi->trapTaskIn.trapBufferBase;
		bufferLen=  tsi->trapTaskIn.trapBufferLength;

		trapTempOut.marker= 0xdfdfdfdf;
		trapTempOut.tFlag= eventType & (1<<30);

		eventAnalysis = eventType & (1<<31);
		if (eventAnalysis) {
			++nEventsTrapped;

			dataPtr1= dataPtr2= bufferBase -0xa00;

			compLen= 0;
			while ((dataPtr2<bufferBase)&&(compLen==0)) {
				if (*dataPtr2 == 0xe0f00000) {    /* find the end-of-fragment */
					compLen= dataPtr2 -dataPtr1 +1;
				}
				++dataPtr2;
			}

			sprintf(genStr,"dataPtr1, *dataPtr1, compLen= %08x, %08x, %08x.\n",
			                dataPtr1, *dataPtr1, compLen);
			newInformation(__FILE__, __LINE__, genStr);
		}


		/* state= TRAP_REQUEST; */
		state= TRAP_EVENTS;
	}


#ifdef COMMENTED
	TRAP_REQUEST:

	++iteration;
	if (init) {
		if (iteration == 1) {
		/* check to make sure that there is a spare trap register, or that one
		   of the registers has been pre-set to trap events. */

		/* send the trap type & event type to the master DSP via the trap request
		   register, and wait for an acknowledgement; make 1st request if not
		   collecting forever or via host. If more than 10 iterations go by before
		   a reply is recieved, the routine assumes the monitoring task is not
		   running and returns an error. */	

		/* newInfo: input parameters/genstr */
		}
		else if ((iteration>5)&&(iteration <= 10)) { /* simulate delay */
			init= 0;
			/*if no reply: return returnCode;
			  else : init= 0, make 1st request & set lastRequestIter */
		}
		else if (iteration> 10) { /* error message & exit w/error */
			return -1;
		}
	}
	if (init) return returnCode;
#endif
 
	if (state == TRAP_EVENTS) {
		if (first) {
			first= FALSE;
			newInformation(__FILE__, __LINE__, "state -> TRAP_EVENTS \n\n.");
			setTaskState(TRAP_TASK, state);
		}

		if ((data= getNextEvent())!= NULL) {
			++eventCount;

/*
			//nErrors= data->nErrors;
			routerError= data->routerError;
			trapEvent= (  ((eventType & 1)&&((nErrors>0)||(routerError>0)))
			            ||((eventType & 2)&&((nErrors==0)&&(routerError==0))) );
*/
			trapEvent= TRUE;

/*
			te1= (eventType & 1);
			te2= (eventType & 2);
			
			te3= ((nErrors>0)||(routerError>0));
			te4= ((nErrors==0)&&(routerError==0));

			sprintf(genStr, "%s %d %d %d %d %d %d %d %d %s",
			"eventType, te1, te2, nErrors, routerError, trapEvent, te3, te4: ",
			eventType, te1, te2, nErrors, routerError, trapEvent, te3, te4, ".\n");
			newInformation(__FILE__, __LINE__, genStr);
*/
			if (eventAnalysis) {
				/* primitive compare: will work for 1 frame & any events which do not
				   cause a buffer overflow or wrap in the middle of the event. */
				dataPtr1= bufferBase -0xa00;
				dataPtr2= (UINT32 *) getFramePtr(data->startFrame);
				for (i=0; i<compLen; ++i, ++dataPtr1, ++dataPtr2) {
					if (*dataPtr1 != *dataPtr2) {
						trapEvent= TRUE;
						sprintf(genStr," Event %07d: Compare failed at word 0x%04x.\n",
							eventCount, i);
						newInformation(__FILE__, __LINE__, genStr);
						break;
					}
				}
			}

			if (data->length <= (bufferLen -dataLen)) {
	
				if (trapEvent) {
/*
					start= (UINT32 *) getFramePtr(data->startFrame);
					start+= 4;
					*start= eventCount;
*/
					fsrp0_on;

					if (trapTempOut.tFlag) copyEvent(data, bufferBase +dataLen);
					/* copyEvent(data, bufferBase +dataLen); */
		
					fsrp0_off;

		
					dataLen+= data->nFrames*0x100;   ++nEventsTrapped;
		/*			dataLen+= data->length;   ++nEventsTrapped;  */
				}
				//processEvent(data, Q_TRAP);
	
				/* indicate in status register that event we requested was found.*/
			}
	
			/* empty trapping buffer and re-fill; send an error message. */
			else {
				if (trapEvent) {
					sprintf(genStr, "%s%x%s%x%s%x%s", "The event trapping buffer has ",
						 dataLen," words used and ", bufferLen-dataLen," available; \n",
						 /*data->length,*/ data->nFrames*0x100,
						 " is needed to trap the current event. Recycling buffer.\n\n");
					newError(&errorCode, MEMORY_EXCEEDED, ERROR_0, "trapTask",
					         genStr, __FILE__, __LINE__);
	
					dataLen= 0;
	
					if (trapTempOut.tFlag) copyEvent(data, bufferBase +dataLen);
					/* copyEvent(data, bufferBase +dataLen); */
					dataLen+= data->nFrames*0x100;   ++nEventsTrapped;
				}
				//processEvent(data, Q_TRAP);


#ifdef COMMENTED
			/* else if (tsi->trapTaskIn.recycle) {
				check if it fits, then remove enough events from buffer to make
				space for them. Three possible values for flag:
				TRAP_TASK_NO_RECYCLE -> error
				TRAP_TASK_RECYCLE_SILENT -> no info message
				TRAP_TASK_RECYCLE_NOTIFY -> info message
	
				once space is available, copy & process event. Re-compute params
			   }   */

			/* A (generally limited) number of trapping attemts will be made. If the
			   iterLimit parameter of the event manager is smaller than the input attempt
			   maximum and it is reached, the event will be treated as a stray event
			   there and we  suddenly will not see it again. Those events may have been
			   sent to the event manager's reserve buffer. */

				++trapAttempts;
				if (trapAttempts == 1) {
					sprintf(genStr, "%s%d%s%d%s%d%s", "The event trapping buffer has ",
						 dataLen," words used and ",
						 (tsi->trapTaskIn.trapBufferLength-dataLen)," available; \n",
						 data->length,
						 " is needed to trap the current event. Event not trapped.\n\n");
				
					newError(&errorCode, MEMORY_EXCEEDED, ERROR_0, "trapTask",
					         genStr, __FILE__, __LINE__);
				}
	
				else if (trapAttempts >= 3) {     /* dummy input = 3*/
					sprintf(genStr, "%s%s%d%s", "The event trapping buffer was ",
						 "still not large enough after ",3,
						 " attempts. Event discarded.\n\n");
				
					newError(&errorCode, EVENT_DISCARDED, ERROR_1, "trapTask",
					         genStr, __FILE__, __LINE__);
	
					processEvent(data, Q_TRAP);
					trapAttempts= 0;
				}
#endif

			}   /* buffer full */

			trapTempOut.nTot= eventCount;
			trapTempOut.nTrapped= nEventsTrapped;

		}    /* event has arrived */
	} /* TRAP_EVENTS */
/*
	if not collecting forever, or via host request:
	if iteration > lastRequestIter & there is acknowledgement by master of last request,
	and in turn we have set the status register flag indicating that the event we
	requested was found, make another request. Otherwise we should wait.  

	
*/
	/**************************************************************************/
	if (nEvents != COLLECT_FOREVER) {

/* & not recycling */
		if (nEventsTrapped >= nEvents) {
			returnCode= SUCCESS;
			tso->trapTaskOut.nEvents=     nEventsTrapped;
			tso->trapTaskOut.dataLen=     dataLen;
			tso->trapTaskOut.bufferBase=  bufferBase;
			tso->trapTaskOut.bufferLen=   bufferLen;
		}
	}

/*
	if (taskMgrCtrlPtr->taskIterations%1000000== 0)
		newInformation(__FILE__, __LINE__, "Event trapping task running.\n");
*/

	return returnCode;
}

/************************************************************************************
 * occupancyTask
 *   
 *   synopsis: Occupancy map task. 
 * 
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 occupancyTask(TaskInput *tsi, TaskOutput *tso,
                    UINT32 opFlags) {
	static INT32  returnCode;
	static UINT8  state= TASK_INIT, first;
	struct EventData *data;	

	/********************* data retrieval & routine reset *********************/
	if (opFlags > 0x10) {
		/* Place the interesting internal variables into the output structure */
		//tso->occupancyTaskOut.nEvents=     nEvents;

		if (opFlags & TASK_HALT_MASK) {
			newInformation(__FILE__, __LINE__, "Occupancy task halting.\n");
			state= TASK_INIT;
			return TASK_HALTED;
		}
		else if (opFlags & TASK_INIT_MASK) {
			newInformation(__FILE__, __LINE__, "Occupancy task re-initing.\n");
			state= TASK_INIT;
			return TASK_INITIALIZED;
		}
		else if (opFlags & TASK_QUERY_MASK) {
			newInformation(__FILE__, __LINE__, "Occupancy task query data output.\n");
			return TASK_QUERIED;
		}
	}
	/*************************** routine states:INIT **************************/
	if (state == TASK_INIT) {
		returnCode= REPEAT_TASK;
	}
	
	/**************************************************************************/
	data= getNextEvent();
	if (data != NULL) {
		data->data2^= (1<<((EVT_MGR_TRAP_O)+(EVT_MGR_TRAP_FXN_OCCUPANCY)) );
	}

	return returnCode;
}

/************************************************************************************/
#pragma CODE_SECTION(decode_data,     "xcode");
#define LINK_UNKNOWN MAX_LINKS //was 0xff TNV

#if defined(PIXEL_ROD)

#else

/********************** used for SCT  decoding ********************************************/
void decode_data(UINT32 word, ErrorInfo *errorInfo){
	// Decode the Router output assuming the SCT Format
  #define rERROR     0x0
  #define rHEAD      0x1
  #define rTRAIL     0x2
  #define rRAW       0x3

  #ifdef OLD_DECODING
  // TODO - add to errorInfo Struct:  //TNV added these.
  // UINT32 errorInfo->theLink 
  // UINT32 errorInfo->theFormat 
  // UINT32 errorInfo->theChip 
  // UINT32 errorInfo->theChannel
  // UINT32 errorInfo->theError;

  // TODO - what happens if LINK_INVALID?  Out of bounds access checks?

  UINT32 halfWord, mask, leftShift;
  static UINT32  enRawErrCnt = 0;


  // Process the two half words in the present data word
  // extract halfword using _extu(word, shiftLeft, shiftRight)
  //  high halfword first,  shiftLeft =  0
  //  high halfword second, shiftLeft = 16

  for (leftShift=0; leftShift<17; leftShift+=16){

  	halfWord = _extu(word, leftShift, 16);  // extract halfWord
    mask     = _extu(halfWord, 16, 29);     // extract the 3 most significant bits of halfWord
  
  	switch(mask) {

  	case rERROR:
  		// Front End Error
        // Word Content: 000x xxxx xFFF FEEE
        //  x = do not care
  		//	F = front end number
  		//  E = ABCD error code

        errorInfo->theChip    = _extu(halfWord, 25,28);
        errorInfo->theChannel = 0;
        errorInfo->theError   = _extu(halfWord, 29,29);

        enRawErrCnt = 0;
        errorInfo->linkErrorCnt[errorInfo->theLink] += 1;
        sprintf(genStr,"Link %d Chip %d Chan %d ABCD_ERROR: code %d: n_errors %d", 
        errorInfo->theLink, errorInfo->theChip, 
        errorInfo->theChannel, errorInfo->theError, errorInfo->linkErrorCnt[errorInfo->theLink]);
        newInformation(__FILE__, __LINE__, genStr);
    	break;

  	case rHEAD:
  		// Header
        // Word Content: 001p tlbK xMMM MMMM (Router output)
        //  p = preamble error, mask 0x1000 
        //  t = time out error, mask 0x0800
        //  l = L1 error ,      mask 0x0400
        //  b = BCID error,     mask 0x0200
        //  K = condensed mode, mask 0x0100
        //  M = link number,    mask 0x007f
        //  all errors          mask 0x1e00

  		errorInfo->theFormat  = _extu(halfWord, 23, 31); // extract bit 8 (i.e., condensed mode)
  		errorInfo->theLink    = _extu(halfWord, 25, 25); // extract lowest 7 bits (i.e., link number)
        errorInfo->theChip    = 0;
        errorInfo->theChannel = 0;
        errorInfo->theError   = 0;

  		enRawErrCnt = 1; // flag to enable error count on first error, flag reset after error

        if(halfWord & 0x1e00){ // any error
            sprintf(genStr,"Link %d HEADER_ERROR:", errorInfo->theLink);
	
  		    if (halfWord & 0x1000){
  		    	errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on preamble error
   	  	    	sprintf(genStr,"%s preamble:",genStr);
		    }	
  		    if (halfWord & 0x0800){
  		    	errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on time out error
   	  	    	sprintf(genStr,"%s timeout:",genStr);
		    }	  		
  		    if (halfWord & 0x0400){
  		    	errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on L1 error
   	  	    	sprintf(genStr,"%s L1error:",genStr);
		    }
  		    if (halfWord & 0x0200){
  		    	errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on BCID error  		
   	  	    	sprintf(genStr,"%s BCerror:", genStr);
		    }
	
            sprintf(genStr,"%s n_errors %d\n", genStr, errorInfo->linkErrorCnt[errorInfo->theLink] );
			newInformation(__FILE__, __LINE__, genStr);
        }
    	
    	break;

  	case rTRAIL:
  		// Trailer
        // Word Content: 010z hvxx xxxx xxxx
        //  z = trailer bit error,          mask 0x1000
        //  h = header trailer limit error, mask 0x0800
        //  v = data overflow error,        mask 0x0400
        //  all errors                      mask 0x1c00

        if(halfWord & 0x1c00){ // any error
            sprintf(genStr,"Link %d TRAILER_ERROR:", errorInfo->theLink);
 	  
  		    if(halfWord & 0x1000){
  			    errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on trailer error
   			    sprintf(genStr,"%s trailer error:",genStr);
            }
  		    if(halfWord & 0x0800){
  			    errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on header-trailer limit error
   			    sprintf(genStr,"%s header trailer limit error:",genStr);
		    }
  		    if(halfWord & 0x0400){
  			    errorInfo->linkErrorCnt[errorInfo->theLink] += 1; // increment on data overflow error
   			    sprintf(genStr,"%s data overflow error:", genStr);
		    }
		
            sprintf(genStr,"%s n_errors %d\n", errorInfo->linkErrorCnt[errorInfo->theLink] );
			newInformation(__FILE__, __LINE__, genStr);
        }
		
        errorInfo->theFormat  = 0;
		errorInfo->theLink    = LINK_UNKNOWN;				
        errorInfo->theChip    = 0;
        errorInfo->theChannel = 0;
        errorInfo->theError   = 0;

   		break;

  	case rRAW:
  		// Raw Data
        // Word Content: 011n nnxx WWWW WWWW
        //  n = count (of raw data)
        //  x = do not care
        //  W = raw data

  		// on RAW the formater has changed because there is a sync bit error
  		// no effort to decode RAW data is done 
  		if(enRawErrCnt){
  			errorInfo->linkErrorCnt[errorInfo->theLink] += 1;
  			enRawErrCnt = 0;
            sprintf(genStr,"Link %d Chip %d Chan %d RAW_ERROR: n_errors %d", 
                errorInfo->theLink, errorInfo->theChip, errorInfo->theChan, errorInfo->linkErrorCnt[errorInfo->theLink]);
			newInformation(__FILE__, __LINE__, genStr);
		}				
    	break;

  	default:
        // Hit Data
        //
        // CONDENSED MODE - single hit
        // Word Content: 1FFF FCCC CCCC xfx0
        //  F = front end number (chip)
        //  C = cluster base address (channel)
        //  x = do not care
        //  f = error in first hit 
        //
        // CONDENSED MODE - double hit
        // Word Content: 1FFF FCCC CCCC sfx1
        //  F = front end number (chip)
        //  C = cluster base address (channel)
        //  s = error in second hit 
        //  f = error in first hit 
        //  x = do not care
        //
        // EXPANDED MODE - 1st hit cluster
        // Word Content: 1FFF FCCC CCCC 0DDD 
        //  F = front end number (chip)
        //  C = cluster base address (channel)
        //
        // EXPANDED MODE - single hit cluster
        // Word Content: 1xxx xxxx 0xxx 1DDD 
        //  x = do not care
        //  D = 3 bit hit data
        //
        // EXPANDED MODE - double hit cluster
        // Word Content: 1xxx xxxx 1DDD 1DDD 
        //  x = do not care
        //  D = 3 bit hit data

        if(errorInfo->theFormat){ // CONDENSED MODE expected
          errorInfo->theChip    = _extu(halfWord, 17,28);  // << 16+1, >> 17+11
          errorInfo->theChannel = _extu(halfWord, 21,25);  // << 16+5, >> 21+4

          if(halfWord & 0x0001){ // double hit
            if(halfWord & 0x000c){ // error in either hit
              sprintf(genStr,"Link %d Chip %d Chan %d HIT_ERROR:", 
                errorInfo->theLink, errorInfo->theChip, errorInfo->theChannel);
              if(halfWord & 0x0004){ // error in first hit
			  errorInfo->linkErrorCnt[errorInfo->theLink] += 1;
   			    sprintf(genStr,"%s first hit:",genStr);
              }
              if(halfWord & 0x0008){ // error in second hit
			  errorInfo->linkErrorCnt[errorInfo->theLink] += 1;
   			    sprintf(genStr,"%s second hit:",genStr);
              }
              sprintf(genStr,"%s n_errors %d\n", genStr, errorInfo->linkErrorCnt[errorInfo->theLink] );
			  newInformation(__FILE__, __LINE__, genStr);
            }
          }else{                 // single hit
            if(halfWord & 0x0004){ // error in first hit
			  errorInfo->linkErrorCnt[errorInfo->theLink] += 1;
              sprintf(genStr,"Link %d Chip %d Chan %d HIT_ERROR:", 
                errorInfo->theLink, errorInfo->theChip, errorInfo->theChannel);
   			  sprintf(genStr,"%s first hit:",genStr);
              sprintf(genStr,"%s n_errors %d\n", genStr, errorInfo->linkErrorCnt[errorInfo->theLink] );
			  newInformation(__FILE__, __LINE__, genStr);
            }
          } 
        }else{ // EXPANDED MODE expected
          if((halfWord & 0x0008) == 0){ // new cluster
            errorInfo->theChip    = _extu(halfWord, 17,28);  // << 16+1, >> 17+11
            errorInfo->theChannel = _extu(halfWord, 21,25);  // << 16+5, >> 21+4
          }
        }

    	break;
    
  	} // end switch
   
   	
  } //end for over half words

  // void - no return code

#endif
}
 


/*
//UINT8 decode_data(UINT32 word, UINT8 link, ErrorInfo *errorInfo){ //TNV
UINT8 decode_data(UINT32 word, UINT8 link, ErrorInfo *errorInfo){

  #define rHEAD      0x1
  #define rTRAIL     0x2
  #define rRAW       0x3
  #define rHIT       0x4 
  #define rERROR     0x0

  int mask;
  int i;
  UINT32 inword, halfWord;
  UINT8 currentlink;
  static UINT32  enRawErrCnt = 0;

  currentlink = link;

  inword = word;
  for (i=0; i<=1; i++){       // process the two half words in word
  	if( i==0) halfWord = (word >> 16) & 0xffff;
	if(i==1)  halfWord = word & 0xffff;
	// debug print
	sprintf(genStr," decode_data() /n first word i=0 ssecond i=1 i= %1d /n halfWord = %4x \n",
					i, halfWord);
	newInformation(__FILE__, __LINE__, genStr);	
  	// mask first three bits that identify type of word
  	// mask = (word & 0xE0000000) >> 29;
  	mask = (halfWord & 0xE000) >> 13;
  	if( mask & 0x4) mask = 0x4 ; // change mask to bits 100 if bit 3 is set for hits
  	//debug print
  	sprintf(genStr,"for switch mask = %4x \n", halfWord);
	newInformation(__FILE__, __LINE__, genStr);	

  	switch(mask) {
  	case rHEAD:
  			//  header word  bits 001p,tlbK,xMMM,MMMM
  				// p = preamble error, mask 0x1000
  				// t = time out error, mask 0x0800
  				// l = L1 error , mask 0x0400
  				// b = BCID error, mask 0x0200
  				
  		enRawErrCnt = 1; // flag to enable error count on first error, flag reset after error
  		currentlink = halfWord & 0x7f; // get 7 bits of link number 
  		// debug print		
  	    sprintf(genStr,"header case header halfword = %4x \n", halfWord);
  	    
		newInformation(__FILE__, __LINE__, genStr);	
  		if (word & 0x1000){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on preamble error
  			// debug print				
   	  		sprintf(genStr,"preamble error \n currentlink = %2d /n errorTaskOut.link[currentlink] %d \n",
   	  				currentlink, errorInfo->linkErrorCnt[currentlink]);
			newInformation(__FILE__, __LINE__, genStr);
		}	
  		if (word & 0x0800){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on time out error
  			// debug print				
   	  		sprintf(genStr,"time out error \n currentlink = %2d /n errorTaskOut.link[currentlink] %d \n",
   	  		        currentlink, errorInfo->linkErrorCnt[currentlink]);
			newInformation(__FILE__, __LINE__, genStr);
		}	  		
  		if (word & 0x0400){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on L1 error
  			// debug print				
   	  		sprintf(genStr,"L1 error \n currentlink = %2d /n errorTaskOut.link[currentlink] %d \n",
   	  		        currentlink, errorInfo->linkErrorCnt[currentlink]);
			newInformation(__FILE__, __LINE__, genStr);	
		}
  		if (word & 0x0200){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on BCID error  		
  			// debug print				
   	  		sprintf(genStr,"BCID error \n currentlink = %2d /n errorTaskOut.link[currentlink] %d \n",
   	  		        currentlink, errorInfo->linkErrorCnt[currentlink]);
			newInformation(__FILE__, __LINE__, genStr);
		}	
    	
    	break;

  	case rTRAIL:
  			// trailer word bits 010z,hvxx,xxxx,xxxx
  				// z = trailer bit error, mask 0x1000
  				// h = header trailer limit error, mask 0x0800
  				// v = data overflow error, mask 0x0400 
  		if( word & 0x1000){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on trailer error
  			// debug print
   			sprintf(genStr,"trailer error \n current link = %2d \n errorTaskOut.linkErrorCnt[currentlink] = %d \n",
   			    	currentlink, errorInfo->linkErrorCnt[currentlink] );
			newInformation(__FILE__, __LINE__, genStr);
		}
  		if( word & 0x0800){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on trailer error
  			// debug print
   			sprintf(genStr,"hader trailer limit error \n current link = %2d \n errorTaskOut.linkErrorCnt[currentlink] = %d \n",
   			    	currentlink, errorInfo->linkErrorCnt[currentlink] );
			newInformation(__FILE__, __LINE__, genStr);
		}		
  		if( word & 0x0400){
  			errorInfo->linkErrorCnt[currentlink] += 1; // increment on trailer error
  			// debug print
   			sprintf(genStr,"data overflow error \n current link = %2d \n errorTaskOut.linkErrorCnt[currentlink] = %d \n",
   			    	currentlink, errorInfo->linkErrorCnt[currentlink] );
			newInformation(__FILE__, __LINE__, genStr);
		}		
		
		currentlink = LINK_UNKNOWN;				
   		break;

  	case rHIT:
    				
    	break;

  	case rRAW:
  		// on RAW the formater has changed because there is a sync bit error
  		// no effort to decode RAW data is done 
  		if(enRawErrCnt){
  			errorInfo->linkErrorCnt[currentlink] += 1;
  			enRawErrCnt = 0;
    		sprintf(genStr,"RAW error data found\n");
			newInformation(__FILE__, __LINE__, genStr);
		}				
    	break;

  	case rERROR:
  		// flaged error  bits 000x,xxxx,xFFF,FEEE
  		//	FFFF = front end number
  		//  EEE  = ABC error code
    	errorInfo->linkErrorCnt[currentlink] += 1;
  		enRawErrCnt = 0;
    	sprintf(genStr,"front end error data found\n");
		newInformation(__FILE__, __LINE__, genStr);
    	break;

  	default:
    
    	break;
    
  	} // end switch
   
   	
  } //end for over half words
  return currentlink;	
} //decode data
*/ 
#endif



#pragma CODE_SECTION(decode_header,     "xcode");
void decode_header(UINT32 word, int nwords){

	enum {HEADSTART, HEADSIZE, HEADFORMAT, 
		  HEADSOURCE, HEADRUNNR, HEADL1ID, HEADBCID, 
		  HEADTTYPE,HEADDTYPE};

	switch(nwords){
		case HEADSIZE:
			sprintf(genStr, "Header: SIZE: %d\n", word);
			break;
		case HEADFORMAT:
			sprintf(genStr, "Header: FORMAT %x\n", word);
			break;
		case HEADSOURCE:
			sprintf(genStr, "Header: SOURCE: %x\n", word);
			break;
		case HEADRUNNR:
			sprintf(genStr, "Header: RUN: %d\n", word);
			break;
		case HEADL1ID:
			sprintf(genStr, "Header: L1ID %d\n", word);
			break;
		case HEADBCID:
			sprintf(genStr, "Header: BCID %d\n", word);
			break;
		case HEADTTYPE:
			sprintf(genStr, "Header: Trigger Type %x\n", word);
			break;
		case HEADDTYPE:
			sprintf(genStr, "Header: DET type %x\n", word);
			break;
		default:
			break;
	}	
	newInformation(__FILE__, __LINE__, genStr);				

}

#pragma CODE_SECTION(errorDetails,     "xcode");
//void errorDetails(struct EventData *data, UINT32 startFrame, UINT32 nDataFrames, 
//				  UINT32 eventwordcount, struct ErrorInfo *errorInfo){
void errorDetails(struct EventData *data, UINT32 startFrame, UINT32 nDataFrames, 
				  UINT32 eventwordcount, ErrorInfo *errorInfo){

    #define mBOF       0xB0F00000
    #define mEOF       0xE0F00000
    #define mHEADSTART 0xEE1234EE


    // need a better place for format definitions
    #define NHEADER     9   
    #define NTRAILER    5
    #define ERROR_COUNT_OFFSET 0  //put this here just to satisfy the compiler TNV

    typedef enum {BOF, HEADSTART, HEADER, DATA, TRAILERSTART, TRAILER, TRAILEREND, EOFS } STATES;

	UINT32 i;
	UINT32 dataCnt, nheaderwords, ntrailerwords;
	STATES state, oldstate, requires;		
	register UINT32 *dataPtr, *startPtr, *endPtr;
	//UINT8  frame, link;  //TNV
	UINT8 frame;
    //UINT32 link; //PWP

	sprintf(genStr, "Error task starting DETAIL nFrames %d nwords %d\n",
			nDataFrames, eventwordcount);
	newInformation(__FILE__, __LINE__, genStr);				

	dataCnt = 0;
	//link    = LINK_UNKNOWN; //PWP
	//errorInfo->theLink = LINK_UNKNOWN; //PWP	
	frame   = startFrame;
	
	/* prepare state machine for BOF*/
	oldstate = EOFS;
	
	for(i = 0; i < nDataFrames; i++) {
		startPtr = (UINT32 *) getFramePtr(frame);
		endPtr   = startPtr + 0x100;  //end of Frame
		dataPtr  = startPtr;
		
		while ((dataPtr < endPtr) && (dataCnt < eventwordcount)) {

			/* Check for marker words */
			if     ((mBOF       == ((*dataPtr) & 0xFFFF0000))) state = BOF;
			else if((mHEADSTART == (*dataPtr))) {
				state          = HEADSTART;
				nheaderwords   = 0;
			} else if((mEOF     == ((*dataPtr) & 0xFFFF0000))) state = EOFS;
			
			if (ntrailerwords==NTRAILER){
				state = TRAILEREND;
				ntrailerwords++;
			}
			
			/* check if Trailer reached */
		    if(dataCnt == (eventwordcount - ERROR_COUNT_OFFSET)) {
		    	 state         = TRAILERSTART;
		    	 ntrailerwords = 0 ;	
			}

	     	/* state machinery to check on proper order of
			   BOF->HEADER->DATA->TRAILER->TRAILEREND->EOF */
			switch(state){
				case BOF:
					requires = EOFS;
 					break;
				case HEADSTART:
					requires = BOF;
					nheaderwords++;
					state = HEADER;	
					break;
				case HEADER:
					requires = HEADER;
					/* decode header words one at a time, keep track of position */
					decode_header((*dataPtr), nheaderwords);
					nheaderwords++;	
					/* DATA should follow after Event Header */
					if (nheaderwords==NHEADER) state = DATA;
					break;
				case DATA:
					requires = DATA;
					/* link comes from trailer info in data */
//					link = decode_data((*dataPtr), link, errorInfo);  //TNV
					decode_data((*dataPtr), errorInfo); // link not used PWP
					break;
				case TRAILERSTART:
					sprintf(genStr, "Error task Trailerstart words: %d cnt: %d tr: %d\n",
							eventwordcount, dataCnt, ntrailerwords);
		  			newInformation(__FILE__, __LINE__, genStr);				
					requires = DATA;
					ntrailerwords++;
					state = TRAILER;		
					break;
				case TRAILER:
					requires = TRAILER;
					ntrailerwords++;	
					break;
				case EOF:
					/* EOF should follow after Event Trailer */
					requires = TRAILEREND;
					break;
			
				default:
					break;
			};
			
			sprintf(genStr, "Error task EVENT word %x count %d state %d\n", (*dataPtr), dataCnt, state);
			newInformation(__FILE__, __LINE__, genStr);				

			/* check state machine */	
			if(requires != oldstate){
				sprintf(genStr, "Error task malformed event words: %d cnt: %d old: %d new: %d reqires: %d\n", 
						eventwordcount, dataCnt, oldstate, state, requires);
	  			newInformation(__FILE__, __LINE__, genStr);				
			}		
			oldstate = state;

			dataPtr++;
			dataCnt++;
			
		}
		frame = getNextFrame(data, frame, TRAP_FXN_ERRORCNT);
	}
	
}

#define ERROR_COUNTING 2
#define ERROR_SUMM_OFFSET 0 //TNV
/******************************************************************************
 * errorTask
 *   
 *   synopsis: error counting & processing task. 
 * 
 *   author: Douglas Ferguson, A.Korn, T. Vickey, P.W. Phillips
 *****************************************************************************/
INT32 errorTask(TaskInput *tsi, TaskOutput *tso, UINT32 opFlags) {
	static INT32  returnCode;
	static UINT8  state= TASK_INIT, first;
	//struct ErrorInfo *errorInfo; //TNV

	INT8 i;
	UINT32 lastframe, nexttolastframe, iframestart, iframecount, 
		xframestart, xframecount, startFrame;
	UINT32 nDataFrames, wordsinlastframe, eventwordcount, endOfBlock;
	UINT32 errorCountWord, errorSummWord;	
	struct EventData *data;

	/********************* data retrieval & routine reset *********************/
	if (opFlags > 0x10) {
		// Place the interesting internal variables into the output structure
		
//		copyMem(&errorInfo,  &tso->errorTaskOut.errorInfo, SIZEOF(errorInfo) );
		tso->errorTaskOut.nErrEvents = errorInfo->nErrEvents;
		tso->errorTaskOut.nErrors    = errorInfo->nErrors;

	    
		if (opFlags & TASK_HALT_MASK) {
			newInformation(__FILE__, __LINE__, "Error Task halting.\n");
			state= TASK_INIT;
			return TASK_HALTED;
		}
		else if (opFlags & TASK_INIT_MASK) {
			newInformation(__FILE__, __LINE__, "Error Task re-initing.\n");
			state= TASK_INIT;
			return TASK_INITIALIZED;
		}
		else if (opFlags & TASK_QUERY_MASK) {
			newInformation(__FILE__, __LINE__, "Error Task query data output.\n");
			return TASK_QUERIED;
		}
	} //opFlags > 0x10
	 
	 
	switch(state){
	/*************************** routine state:INIT *********************/
		case TASK_INIT:
			// Initialize both the local and task output errorInfo structures 
			setMem((UINT32 *) &errorInfo,   SIZEOF(errorInfo), 0);
			setMem((UINT32 *) &tso->errorTaskOut, SIZEOF(tso->errorTaskOut), 0);

			state      = ERROR_COUNTING;
			returnCode = REPEAT_TASK;
			break;	


	/**************************** routine state: ERROR_COUNTING ********/
		case ERROR_COUNTING:
			data = getNextEvent();
			if (data == NULL) {
				return returnCode;
        	}

			// Error decoder is now in histogram_sct.c
         //   histoError_c(data);

#ifdef 0	
		    // Letting decode_data() in errorDetails do most of the counting
			//errorInfo.eventCount++;	
			tso->errorTaskOut.errorInfo.errorCount++;
			
			// Clear trapped event type ???
	 		data->data2   ^= (1<<((EVT_MGR_TRAP_O)+(EVT_MGR_TRAP_FXN_ERRORCNT)) );

	  		// Decode last and next to last frame from data3
	 		lastframe       = ((data->data3) & 0xFF000000) >> EVT_MGR_CFRAME_O;
	 		nexttolastframe = ((data->data3) & 0xFF0000)   >> EVT_MGR_LFRAME_O;

	  		// Fast check for errors, last word in last frame:
	  		// EFB errors??? Independent of PXL or SCT??? TNV
	  		//  01000000000t11dewwwwwwwwwwwwwwww
	  		//  d bit 17: NOT event data loss, the event error is not set
	  		//  e bit 16: NOT event error, excludes data lose
	  		//  t bit20 : trap number 
	  		// 	
	 		endOfBlock      = *((UINT32 *)getFramePtr(lastframe) + 0x100 -1);
      		// Leave if no error
	  		if(1==((endOfBlock >> 16) & 1)) return returnCode;
	  					       
	  		// Handle data loss ...
	  		if(0==((endOfBlock >> 17) & 1)) {			       
	   	 		//tso->errorTaskOut.errorInfo.type[ERRORTYPE_DATALOSS]++;
	    		return returnCode;
        	}
	
	  		// Decode frame pointers in data1 
	  		iframestart = (data->data1)              >> EVT_MGR_IFRAME_O;
	  		iframecount = ((data->data1) & 0xFF0000) >> EVT_MGR_IFCNT_O;
	 		xframestart = ((data->data1) & 0xFF00)   >> EVT_MGR_XFRAME_O;
	  		xframecount = ((data->data1) & 0xFF);

	  		nDataFrames    = iframecount + xframecount;
	  		eventwordcount = (endOfBlock & 0xFFFF);

	  		// Event words in last frame
	  		wordsinlastframe = eventwordcount-((nDataFrames-1)*0x100);

	  		// Frame location of error counter word
	  		if (wordsinlastframe > ERROR_COUNT_OFFSET){
	    		// in last frame 
	    		errorCountWord = *((UINT32 *)getFramePtr(lastframe) + 
			 	      (wordsinlastframe - 1 - ERROR_COUNT_OFFSET));
	  		} else {
	    		// ...in next to last frame
	    		errorCountWord = *((UINT32 *)getFramePtr(nexttolastframe) + 
					       (0x100-ERROR_COUNT_OFFSET + wordsinlastframe - 1));
	  		}

	  		// Leave if not an error event
	  		if (0 == errorCountWord) return returnCode;
      		//AKDEBUG
	    	// Last words in last frame
	    	sprintf(genStr, "Error Task %s%x\n %s%x\n %s%x\n %s%x\n %s%x\n %s%x\n",
	    	"word[0]: ", *((UINT32 *)getFramePtr(lastframe) + wordsinlastframe - 6),
	    	"word[1]: ", *((UINT32 *)getFramePtr(lastframe) + wordsinlastframe - 5),
	    	"word[2]: ", *((UINT32 *)getFramePtr(lastframe) + wordsinlastframe - 4),
	    	"word[3]: ", *((UINT32 *)getFramePtr(lastframe) + wordsinlastframe - 3),
	    	"word[4]: ", *((UINT32 *)getFramePtr(lastframe) + wordsinlastframe - 2),
	    	"word[5]: ", *((UINT32 *)getFramePtr(lastframe) + wordsinlastframe - 1));
	  		newInformation(__FILE__, __LINE__, genStr);
			sprintf(genStr, "Error task gotten error cnt: %x  words: %x\n", 
				    errorCountWord, wordsinlastframe);
	  		newInformation(__FILE__, __LINE__, genStr);

	  		//errorInfo.errorEventCount++;
	  		tso->errorTaskOut.errorInfo.errorEventCount++;
	  		//errorInfo.errorCount += errorCountWord;
		    tso->errorTaskOut.errorInfo.errorCount += errorCountWord;

	  		// Frame location of error summary word
	  		if (wordsinlastframe > ERROR_SUMM_OFFSET){
	    		// ...in last frame 
	    		errorSummWord = *((UINT32 *)getFramePtr(lastframe)+ 
					       (wordsinlastframe - 1 - ERROR_SUMM_OFFSET));
	  		} else {
	    		// ...in next to last frame
	    		errorSummWord = *((UINT32 *)getFramePtr(nexttolastframe) + 
					      (0x100 - ERROR_SUMM_OFFSET + wordsinlastframe - 1));
	 	 	}

	        /* Not yet used for SCT.  TNV
	  		// histogram each error bit
	  		for(i=0;i<N_EFB_ERROR_BITS;i++){
	    		//if((errorSummWord & (1<<i)))errorInfo.type[i]++;
	    		if((errorSummWord & (1<<i)))tso->errorTaskOut.errorInfo.type[i]++;
	  		}
		    */
						  
	  		// Start of detailed error counting
	  		startFrame = iframestart;
	  		if (0==iframecount) startFrame = xframestart;
			errorDetails(data, startFrame, nDataFrames, 
					     eventwordcount, errorInfo);	//TNV
//							 eventwordcount, tso->errorTaskOut.errorInfo); //complaints

#endif	
 			break;
		default:
			break;			

	} // end switch
	return returnCode;
}


/************************************************************************************
 * resynchTask
 *   
 *   synopsis: Re-synchronization of events in case of detector electronics glitch. 
 * 
 *   author: Douglas Ferguson
 ************************************************************************************/
INT32 resynchTask(TaskInput *tsi, TaskOutput *tso,
             UINT32 opFlags) {
	static INT32  returnCode;
	static UINT8  state= TASK_INIT, first;
	struct EventData *data;

	/********************* data retrieval & routine reset *********************/
	if (opFlags > 0x10) {
		/* Place the interesting internal variables into the output structure */
		//tso->occupancyTaskOut.nEvents=     nEvents;

		if (opFlags & TASK_HALT_MASK) {
			newInformation(__FILE__, __LINE__, "Resynch task halting.\n");
			state= TASK_INIT;
			return TASK_HALTED;
		}
		else if (opFlags & TASK_INIT_MASK) {
			newInformation(__FILE__, __LINE__, "Resynch task re-initing.\n");
			state= TASK_INIT;
			return TASK_INITIALIZED;
		}
		else if (opFlags & TASK_QUERY_MASK) {
			newInformation(__FILE__, __LINE__, "Resynch task query data output.\n");
			return TASK_QUERIED;
		}
	}
	/*************************** routine states:INIT **************************/
	if (state == TASK_INIT) {
		returnCode= REPEAT_TASK;
	}
	
	/**************************************************************************/
	data= getNextEvent();
	if (data != NULL) {
		data->data2^= (1<<((EVT_MGR_TRAP_O)+(EVT_MGR_TRAP_FXN_RESYNCH)) );
	}
	
	return returnCode;
}
