/**
   File: SctApi.h

   Library interface to SCTAPI
   Defines SctApi as used by TApi and SctApiServer
*/

#ifndef SCTROD_SCTAPI_H
#define SCTROD_SCTAPI_H

// Make sure all compilations have same SctApi structure!
#include "config.h"

#include <list>
#include <map>
#include <string>
#include <vector>

#include <boost/thread.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>

#include <mrs/message.h>
#if USE_IS
//#include <is/isinfotmpl.h>
#include <is/infodictionary.h>
#endif

class SCTDCSCommand;
#include "log.h"

#include "SctApiConfigCache.h"
#include "Sct/Addressing.h"

// This defines ABCDModule (a typedef)
#include <Sct/AbcdModule.h>
// Unfortunately this is necessary
#include <CommonWithDsp/processor.h>

struct scan_result_ptrs;

namespace SctConfiguration {
  class Configuration;
  struct BOCChannelConfig;
  struct BOCGlobalConfig;
}

#include "SctApiFwd.h"

// includes
#include "ModuleList.h"

namespace SctApi {

/** 
    SctApi class.

    This is the API used by the rest of the world to interact with
    SCT Modules using RODs.

    Functionality that should be included in this class includes.

    Modules are referred to by an "mid" which is a 32bit number 
    packed in hex as PPCCRRCH where:
       PP is the partition
       CC is the crate
       RR is the ROD
       CH is the ROD channel

    DAQ run level transitions:
       Boot
           Load and run the crate controller (Done automatically in DAQ-1 from configuration)
       Load
           Load config?
       Config
           Send config to Modules
       Start
           Mode for execution of scans

    Configuration:
      loading from file
      loading to modules
      fine grained control on modules
      recovering module configuration from the ROD
      Save configuration

      Reconfiguring from new configuration file
    
    Scanning:
      perform scans over configuration variable(s)

    DCS:
      minimal DCS communication

    ROD diagnostics:
      Flash LEDs
      Text buffer retrieval
      Memory dumps
      Send arbitrary primitives to a single ROD

    Indexing:
      There are several ways of addressing a ROD or module
      the canonical method is
              partition, crate, rod, module

    Other methods include:
       Geographical:
           barrel/disc MUR, number
       MUR:
           Partition, MUR, number
*/
class SctApi {
  SctApi(const SctApi &);
  SctApi &operator=(const SctApi&);
 public:

  /** The constant returned as an unknown module */ 
  static const unsigned int BAD_MODULE;

  /**
     Default to local xmlConfig
  */
  SctApi(const Idiosyncrasy &); // copies the idiosyncrasy

  /**
     Specify configuration
  */
  SctApi(const Idiosyncrasy &, boost::shared_ptr<SctConfiguration::Configuration> newConf); // copies the idiosyncrasy

  virtual ~SctApi();

 private:
  /**
     This sets the unexpected handler to interpret all sorts of exceptions

     Called by constructor
   */
  void setup();

  /**
     If requested, dump prim list to file and/or log to output
   */
  void dumpPrimList(boost::shared_ptr<PrimListWrapper> prim);

  /// Save raw event to file
  void dumpRawEvent(unsigned int rod, int units, unsigned long *bufferA, unsigned long *bufferB);

public:
  /**
     Method to allow others to see what makes us "special" in comparison to allthe other crate controllers
  */
  const Idiosyncrasy & idiosyncrasy();
  const Sct::UCID & ucid() const; // shortcut to m_id->ucid();

  const std::string getHostName() const;

  /**
     C++ only method to tell a MRS aware program where to write to
   */
  void setMrsStream(MRSStream *stream);

#if USE_IS
  /**
     C++ only method to tell a IS aware program to go and look up the dictionary which will tell it where to write to
   */
  void setIsDictionary();
#endif

  /***************** Global configuration functions ***********************/

  /// Initialise all software components
  /** Initialise the crates and create the Rod/Vme objects.

      Load the system configuration from the database.
      Create crates and RODs as necessary.
      Calls initialiseRod and initialiseModule as necessary.

      Store the module configuration in the RODs (UNIMPLEMENTED)

      Return: 0 for success
              Possibly error code for failure

      IPC: Block everything else!
  */
  void initialiseAll(int run);

  /// Tidy up and shutdown
  /** Leave everything in a suitable state and delete all the objects

      Shutdown anything that needs shutting down
      This is where we should send DCS to standby message??

      IPC: Block everything else!
  */
  void shutdownAll();

  /// Change the run number
  /**
     IPC: Don't do during a scan
  */
  void setRunNumber(UINT32 newRun);

  /// Change the next scan number
  /**
     IPC: Don't do during a scan
  */
  void setScanNumber(UINT32 newScan);

  /// Return the run number
  /**
     IPC: Don't do during a scan
  */
  UINT32 getRunNumber();

  /// Return the number of the next scan
  /**
     IPC: Don't do during a scan
  */
  UINT32 getScanNumber();

  /// Check if a ROD has been configured
  /**
     Return true if ROD can accept commands
     (Doesn't go so far as to access ROD hardware)
   */
  bool isRODPresent(unsigned int rod);

  /**
     Change to physics/calibration mode
   */
  bool changeRunMode(enum RunType mode);

  /**
     Change the event format produced by the formatters
   */
  bool changeEventFormat(enum EventFormatType format);

  /// Send the configurations from the RODs to the modules.
  /**
     DDC: Check modules in configurable state.
          Correct run state and no warnings

     IPC:  Don't do during a scan
     Conf: Maybe interaction
   */
  void startupModules();

  /// Return list of configured modules
  ModuleList getModuleList();

  /*********************** Module configuration **********************/

  /// (Re)Load configuration from the "configuration database"
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void loadConfiguration();

  /// Load new configuration from a "configuration database"
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void loadConfiguration(std::string filename);

  /// Load module configurations from database
  /**
     Return number of module loaded
   */
  int loadModuleConfigurations();

  /// Store the module configurations of a bank into database
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void storeModuleConfigurations(BankType bank);

  /// Store the current BOC configurations into database
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void storeBOCConfigurations();

  /// Reset BOC configuration from current 
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void configureBOC(unsigned int rod);

  /// Get array of BOC monitor values
  /**
     IPC:  Should be OK during scan, only reads from BOC
     Conf: Nothing
   */
  std::vector<double> getBOCMonitorArray(unsigned int rod);

  /// Wait for a ROD to finish a primlist
  /**
     IPC:  Allow during a scan (only expert mode)
     Conf: no interaction!
  */
  int awaitResponse(unsigned int rod, int timeout);

  /// Wait for all RODs in a crate to finish a primlist
  /**
     IPC:  Allow during a scan (only expert mode)
     Conf: no interaction!
  */
  int awaitResponseAll(int timeout);

  /// Return the out list
  /**
     (No interaction with VME)
     IPC:  Allow during a scan (only expert mode)
     Conf: No interaction!
  */
  unsigned long *getResponse(unsigned int rod, 
                             unsigned long &length);

  /** Get a message in the text buffer

      Get a message from a ROD text buffer
      Returns whether the action was completed
  */
  bool getRodMessage(unsigned int rod,
                     char *buffer, unsigned long &length);

  /*************** ROD Diagnostics ********************/

  /// Flash the LED associated with a slave DSP. 
  /**
     This creates its own primitive list and sends it to 
     a slave DSP on the ROD.

     @param period Flash period given in ms
     @param slaveNumber Slave number or -1 for Master

     IPC:  Possible during a scan?
     Conf: No interaction
  */
  void flashLED(unsigned int rod,
                long slaveNumber, long period = 1000, long flashes = 10);

  /// Echo data to a ROD
  /**
     IPC:  Possible during a scan?
     Conf: No interaction
  */
  void echo(unsigned int rod,
            unsigned int length, const unsigned long *data);

  /// Echo data to all RODs
  /**
     IPC:  Possible during a scan?
     Conf: No interaction
  */
  void echoAll(unsigned int length, const unsigned long *data);

  /// Echo data to a Slave in a ROD
  /**
     IPC:  Possible during a scan?
     Conf: No interaction
  */
  void echoSlave(unsigned int rod,
                 unsigned int slave, unsigned int length, const unsigned long *data);


  /***************** Direct ROD access (Debug only) ********************/
  /* Should be able to do anything with these commands */

  /// Dump a block of memory direct from a DSP
  /**
     Dump contents of a block of DSP memory to cout.

     @param dspNumber Slave number, -1 for the Master DSP.
     @param usePrim Use a primitive instead of direct VME access

     IPC:  Possible during a scan
     Conf: No interaction
  */
  int dspBlockDump(unsigned int rod,
                   long dspStart, long numWords, long dspNumber, bool usePrim = true);

  /// Dump a block of memory direct from a DSP to a file
  /**
     Dump contents of a block of DSP memory to a file.
     Use dspNumber -1 to refer to the master DSP.

     @param dspNumber Slave number, -1 for the Master DSP.
     @param usePrim Use a primitive instead of direct VME access

     IPC:  Possible during a scan
     Conf: No interaction
  */
  int dspBlockDumpFile(unsigned int rod,
                       long dspStart, long numWords, long dspNumber, std::string filename, bool usePrim = true);

  /// Perform a dump of the ROD with standard file names
  /**
     Dump to ${tempDir}/Dump_%05d

     @param rl Refers to the ROD to dump
   */
  void standardRegisterDump(RodLabel rl);

  /// Perform a dump of the ROD with standard file names
  /**
     Dump to ${tempDir}/Dump_%05d
   */
  void standardRegisterDump(unsigned int rod);

  /// Dump registers of all RODs to files
  /**
     Dump to ${tempDir}/Dump_%05d
     RODs sent to consecutive %d
   */
  void standardRegisterDumpAll();

  /// Read a block of memory direct from a DSP
  /**
     Use dspNumber -1 to refer to the master DSP.

     IPC:  Possible during a scan
     Conf: No interaction
  */
  unsigned long *dspBlockRead(unsigned int rod,
                              long dspStart, long numWords, long dspNumber, unsigned long &length, bool usePrim = true);

  /// Write a block direct to a DSP
  /**
     Use dspNumber -1 to refer to the master DSP.

     IPC:  Possible during a scan
     Conf: No interaction
  */
  int dspBlockWrite(unsigned int rod,
                    unsigned long *buffer, unsigned long dspAddress, long numWords, long dspNumber, bool usePrim = true);

  /// Read single word to DSP using block read
  /**
     Same prototype as dspSingleRead but safer (for slave DSPs)
   */
  unsigned long dspSingleBlockRead(unsigned int rod,
                                   long dspStart, long dspNumber);

  /// Write single word to DSP using block write
  /**
     Same prototype as dspSingleWrite but safer (for slave DSPs)
   */
  void dspSingleBlockWrite(unsigned int rod,
                           unsigned long dspAddress, unsigned long val, long dspNumber);

  /// Read a block of slave DSP memory using RW_SLAVE_MEMORY
  /**
     Don't use if important to preserve primitive state!

     Returns new[]ed block of memory

     IPC:  Possible during a scan
     Conf: No interaction
  */
  unsigned long *primReadSlaveDsp(unsigned int rod,
                                  int s, int add, int words);

  /// Write a block of slave DSP memory using RW_SLAVE_MEMORY
  /**
     Don't use if important to preserve primitive state!

     IPC:  Possible during a scan
     Conf: No interaction
  */
  void primWriteSlaveDsp(unsigned int rod,
                         int s, int add, int numWords, unsigned long *data);

  void writeSlaveFile(unsigned int rod, unsigned int slave, 
                      std::string fileName, unsigned long address, bool usePrim = true);

  /// Send the start slave primitive to a DSP
  void startSlave(unsigned int rod, unsigned int slave);

  /// Get block of memory from file
  boost::shared_array<unsigned long> loadFile(std::string filename, unsigned long & length);

  /// Read one ROD Status Register via HPI
  /**
     IPC:  Immediate response
     Conf: No interaction
  */
  unsigned long readRodStatusReg(unsigned int rod,
                                 long regNumber);

  /// Read one ROD Command Register via HPI
  /**
     IPC:  Immediate response
     Conf: No interaction
  */
  unsigned long readRodCommandReg(unsigned int rod,
                                  long regNumber);

  /// Read a single 32b word from MasterDSP SDRAM via HPI
  /**
     IPC:  Immediate response
     Conf: No interaction
  */
  unsigned long dspSingleRead(unsigned int rod,
                              const unsigned long dspAddr, long dspNumber);

  /// Write a single 32b word to MasterDSP SDRAM via HPI
  /**
     IPC:  Immediate response (expert only)
     Conf: No interaction
  */
  void dspSingleWrite(unsigned int rod,
                      unsigned long dspAddr, unsigned long value, long dspNumber);

  /// Retrieve module configuration data
  /**
     Module data as cached from Configuration

     IPC:  Immediate response
     Conf: Interaction
  */
  ABCDModule *retrieveModule(UINT32 mid);

  /// Create a prim list
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void createDebugPrimList();

  /// Insert a primitive in a prim list
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void addDebugPrimList(unsigned long length, long index, long id, long version,
                        unsigned long * body);

  /// Send the previously created prim list
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void sendDebugPrimList(unsigned int rod);

  /// Send the previously created prim list to all RODs
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void sendDebugPrimListAll();

  /// Send the "Debug" primitive to a slave dsp
  /**
     Send the previously created prim list to a slave dsp
     @param await Master prim list waits for slave list to complete 
     @param response Expect a reply 

     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void sendDebugSlavePrimList(unsigned int rod,
                              unsigned int slave, bool await, bool response);

  /// Load a binary dump of a primitive list into the debug prim list
  /**
     This doesn't check the primlist checksum (but slave lists will need the correct sum)
   */
  void debugPrimListFromFile(std::string fileName);

  /// Perform a binary dump of the debug prim list
  void dumpDebugPrimList();

  /// Return a list of the RODs that have been configured
  const std::list<RodLabel> listRods() const;

  /// Return a list of the RODs that have been configured
  //const std::list<CrateLabel> listCrates() const;

  /**** Official SCTAPI bit ******/

  /// Return unique identifier for a module
  /**
     @param sn Module serial number
     @return The unique identifier, BAD_MODULE if not found

     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findModule(std::string sn);

  /// Return std::string for uid
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  std::string convertToString(UINT32 mid);

  /// Return unique identifier for mur/module 
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findModule(INT32 mur, INT32 module);

  /// Return mur/module pair for uid
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  std::pair<INT32, INT32> convertToMUR(UINT32 mid);

  /// Return unique identifer from barrel geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findBarrelModule(INT32 barrel, INT32 row, INT32 number);

  /// Convert uid to barrel geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  void convertToBarrelModule(UINT32 mid, UINT32 &barrel, UINT32 &row, int &number);

  /// Return unique identifer from endcap geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findEndcapModule(INT32 disk, INT32 quadrant, INT32 number);

  /// Convert uid to endcap geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  void convertToEndcapModule(UINT32 mid, INT32 &disk, UINT32 &quadrant, UINT32 &number);

  /// Get all modules from ROD to memory
  /**
     IPC:  Immediate response
     Conf: Lookup
     LocalCache: Yes
  */
  void getABCDModules(BankType bank);

  /// Get module from ROD to memory
  /**
     IPC:  Immediate response
     Conf: Lookup
     LocalCache: Yes
  */
  virtual void getABCDModule(UINT32 mid, BankType bank);

  /// Get module configuration as stored in ROD, for debugging
  /**
     IPC:  Immediate response
     Conf: Lookup
     LocalCache: No
  */
  virtual boost::shared_ptr<ABCDModule> getABCDModuleRaw(unsigned int rod, 
                                                         UINT32 slot, BankType bank);
  
  /// Set all module configs in a ROD bank
  /**
     IPC:  Don't do during scan
     Conf: Possible cache
     LocalCache: Yes
  */
  virtual void setABCDModules(std::list<BankType> bank);

  /// Set module config in ROD bank
  /**
     IPC:  Don't do during scan
     Conf: Possible cache
     LocalCache: Yes
  */
  virtual void setABCDModule(UINT32 mid, std::list<BankType> banks);

  /** 
      Do internal ROD copy of its banks
   */
  virtual void copyABCDModule(UINT32 mid, BankType from, std::list<BankType> to);

  /** 
      Do internal ROD copy of its banks - all modules
   */
  virtual void copyABCDModules(BankType from, std::list<BankType> to);

  /// Send configuration for all modules from the ROD bank to the module
  /**
     @param bank Configuration bank in the ROD
     @param type How much configuration to transmit

     DDC:  Check module state
           Set select line
     IPC:  Not during scan
     Conf: No Lookup
     LocalCache: not as of 1/3/2005 AJB
  */
  virtual void sendAllABCDModules(BankType bank,
                                  ConfigType type = SCTAPI_CONFIG_ALL,
                                  const bool enableDataMode=true /* enableDataMode should have default value true to ensure compatibility with old code*/);

  /// Send configuration for one module from the ROD bank to the module
  /**
     @param mid Unique identifier of the module to send
     @param bank Configuration bank in the ROD
     @param type How much configuration to transmit

     DDC:  Check module state
           Set select line
     IPC:  Not during scan
     Conf: No Lookup
  */
  virtual void sendABCDModule(UINT32 mid,
                              BankType bank,
                              ConfigType type = SCTAPI_CONFIG_ALL,
                              const bool enableDataMode=true /* enableDataMode should have default value true to ensure compatibility with old code*/);

  /// Modify the mask of a module configuration
  /**
     Set the channel mask of all chips in a module

     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDMask(UINT32 mid, UINT32* mask);

  /**
     Modify the mask of a list of module configuration
  */
/*    void modifyABCDListMask(mList* mlist, UINT32* mask); */

  /**
     Modify a module's trim configuration

     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDTrims(UINT32 mid, UINT8* trims);

  /**
     Modify a module list's trim configuration
  */
/*    void modifyABCDTrims(mList* mlist, UINT8* trims); */

  /** 
      List of banks to be kept up to date with 
  */
  void setAutoUpdateBanks(std::list<BankType> banks);

  /// Modify all module configurations. Will also update on ROD banks listed in autoUpdateBanks.
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVar(UINT32 typ, FLOAT32 var);

  /// Modify variable in module configuration.  Will also update on ROD banks listed in autoUpdateBanks.
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVar(UINT32 mid, UINT32 typ, FLOAT32 var);

  /// Modify variable in list of modules.  Will also update on ROD banks listed in autoUpdateBanks.
/*    void modifyABCDVar(mList* mlist, UINT32 typ, FLOAT32 var); */

  /// Modify variable on one chip of module configuration 
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVar(UINT32 mid, UINT32 chip, UINT32 typ, FLOAT32 var);

  /**
     modify a variable on one chip, but all modules
   */
  void modifyABCDVarAllModulesOneChip(UINT32 chip, UINT32 typ, FLOAT32 var);

  /// Modify variable on one chip of the module configuration on the ROD
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVarROD(UINT32 mid, UINT32 chip, UINT32 typ, FLOAT32 var, std::list<BankType> banks);

  /// Modify variable on one module configuration on the ROD
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVarROD(UINT32 mid, UINT32 typ, FLOAT32 var, std::list<BankType> banks);

  /// Modify variable of all module's configuration on the ROD
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVarROD(UINT32 type, FLOAT32 value, std::list<BankType> banks);

  /// Modify variable on one chip of module configurations in a list
/*    void modifyABCDVar(mList* mlist, UINT32 chip, UINT32 typ, FLOAT32 var); */

  /// Modify response curve setting on one chip
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDRC(UINT32 mid, UINT32 chip, UINT16 function, FLOAT32 p0, FLOAT32 p1, FLOAT32 p2);

  
  /**
     modify a TIM parameter.
   */
  void modifyTIMParam(unsigned int param, unsigned int value);


  /// Modify BOC parameter
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyBOCParam(unsigned int rod,
                      unsigned int channel, unsigned int type, unsigned int val);

  /// Modify all BOC parameters
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyBOCParam(unsigned int type, unsigned int val);

  /// Print BOC parameters for all channels
  /**
     IPC:  Immediate response
     Conf: No communication
  */
  void printBOCSetup(unsigned int rod);

  /// Return the current BOC parameters for all channels
  /**
     IPC:  Immediate response (Needs results by IPC)
     Conf: No communication
  */
  std::vector<SctConfiguration::BOCChannelConfig> currentBOCSetup(unsigned int rod);

  /// Print BOC registers (independent of channels)
  /**
     IPC:  Immediate response
     Conf: No communication
  */
  void printBOCRegisters(unsigned int rod);

  /// Return the current BOC global registers
  /**
     IPC:  Immediate response (Needs results by IPC)
     Conf: No communication
  */
  SctConfiguration::BOCGlobalConfig currentBOCRegisters(unsigned int rod);

  /// Save BOC setup
  void saveBOCSetup(unsigned int rod, BankType bank);

  /// Save BOC registers
  void saveBOCRegisters(unsigned int rod, BankType bank);

  /// Response BOC setup
  void restoreBOCSetup(unsigned int rod, BankType bank);

  /// Restore BOC registers
  void restoreBOCRegisters(unsigned int rod, BankType bank);

  /// Turn all the BOC laser currents off (modules lose clock, not a good idea)
  /**
     IPC:  Immediate response
     Conf: No communication
  */
  void lasersOff();

  /// put all EFB error masks to their cached values
  void setErrorMasks();

  /* *************  TIM functs **************/

  /// Set TIM trigger frequency
  /**
     Freq is approximated by divider circuit
     Available frequencies are
        600, 60, 6, 0.6  kHz
     Divided by:
        1, 2, 3, 4, 5, 6, (10, 12)
   */
  void timSetFrequency(double trigFreq, double rstFreq);

  /// get trigger frequency (kHz)
  double timGetTriggerFrequency();

  /// get reset frequency (Hz)
  double timGetResetFrequency();

  /// Start the TIM generating (internal) regular triggers 
  /**

     IPC:  Not during scan
     Conf: No communication
  */
  void freeTriggers();

  /// Stop the TIM generating (internal) regular triggers
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void stopTriggers();

  /// Tell the TIM to send an L1A 
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void timL1A();

  /// Tell TIM to send Cal + L1A
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void timCalL1A(int delay);

  /// Tell TIM to send soft Reset
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void timSoftReset();

  /// Tell TIM to send BC Reset
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void timBCReset();

  /// load a tim sequence - see Crate for detail
  void timLoadSequence(const std::vector<unsigned>& seq);
  
  /// run the tim sequence @return last l1id  
  unsigned long timRunSequence(unsigned long ntrigs, int random, bool external=false);
  
  /// Send burst of triggers to TIM @return last l1id
  unsigned long sendTimBurst(unsigned long count, int random, bool external=false);

  /// Print all TIM registers to stdout
  void timVerbose();

  /// Write a TIM register
  /**
     @param reg VME address of the register
   */
  void timWriteRegister(int reg, UINT16 val);

  /// Read a TIM register
  /**
     @param reg VME address of the register
   */
  UINT16 timReadRegister(int reg);

  /// Print status to cout
  /**
     IPC:  Immediate response (sane in all eventualities!)
     Conf: No communication
  */
  std::string status();

  /// Do a scan (hardcoded for testing)
  /**
     IPC:  Not during any other scan
     Conf: No communication
  */
  void defaultScan(int type);

  /// Coordinate a scan
  /**
     DDC:  Check modules configured...
           Check select
     IPC:  Not during any other scan
     Conf: Lookup
  */
  void doScan(boost::shared_ptr<const Scan> scan);

  /// High level method to start a scan in physics running
  /**
     DDC:  Check modules configured...
           Check select
     IPC:  Not during any other scan
     Conf: Lookup
  */
  boost::shared_ptr<ScanMonitor> startAsyncScan(boost::shared_ptr<const Scan> scan);

  /// Start a scan in physics running
  /**
     DDC:  Check modules configured...
           Check select
     IPC:  Not during any other scan
     Conf: Lookup
  */
  boost::shared_ptr<ScanControl> startScan(boost::shared_ptr<const Scan> scan, bool pollTriggers);

  /** 
      Check that a scan is ok to run
      @throw SctApiException if there is a problem in the scan
   */ 
  void SctApi::validateScan(boost::shared_ptr<const Scan> scan);

  /// Used to determine whether or not SctApi is busy.  See also "SctApi::unsetBusy()".
  bool hasFinishedScanning();

  const std::string & nameOfScanStatusObject();

  /**
     Await completion of a scan (alternative to checking IS running entry)
   */
  void awaitScan();

  /**
     Abort any long term activity

     Unimplemented

     DDC:  No action
     IPC:  Immediate response
     Conf: No communication
   */
  void abortScan();

  /// Scan returning histograms of raw data
  /**
     IPC:  Not during any other scan
     Conf: Lookup
  */
  void doRawScan(boost::shared_ptr<Scan> scan, int delay, int width, 
                 bool configureModules = false, bool clkBy2 = false);

  /// Send one trigger to modules in one ROD
  /**
     IPC:  Not during any other scan
     Conf: No communication
  */
  void sendRodTrigger(unsigned int rod, const Trigger *trig, long unsigned count, int scan_point=0);

  /**
     Send lots of triggers: frequency and source worked out from trigger
     @return last l1id
   */
  unsigned long sendTriggers(const Trigger* trig, unsigned long count, int scan_point=0);

  /// Print the module configuration in memory
  /** 
     Uses the sctConf method

     IPC:  Not very useful without feedback
     Conf: Possible lookup
  */
  void printABCDModule(int mid);

  /// Print the module configuration from the ROD 
  /**
     First retrieve the configuration from the ROD
     Uses the sctConf method

     IPC:  Not very useful without feedback
     Conf: Possible lookup
   */
  void printABCDRodModule(int mid, BankType bank);

  /// Print decoded event
  /**
     IPC:  Needs feedback, immediate response
     Conf: No Communication
  */
  void decodeEvent(unsigned int rod, 
                   int sl, int index, bool extFlag = false, bool errorType = false);

  /// Scan event buffer
  /**
     IPC:  Needs feedback, immediate response
     Conf: No Communication
  */
  void scanEvents(unsigned int rod, int sl, bool extFlag, bool errorType);

  /**
     Suggest problems with modules based on events in event buffer
   */
  void reportEventErrors();

  /**
     Decode config bit stream (only valid after SEND_CONFIG)
   */
  void decodeConfig(unsigned int rod, 
                    bool skipTrim = false, bool bypass = false);

  /// Read raw link data (send L1A to trigger some data)
  /**
     IPC:  Needs feedback, immediate response
     Conf: No Communication
  */
  void rawData(unsigned int rod, int delay, int units, bool setMask = true, const Trigger *trig = 0);

  /// Send an L1A to all modules
  /**
     The mask is unchanged in this function and it is set up to go to all modules

     IPC:  Not during scan
     Conf: No Communication
  */
  void sendL1A(unsigned int rod, bool capture = false);

  /// Make guess at what's attached to different channels
  /**
     Do quick scan of channels
     Recognises all the same, clock by 2 and an event header
     Return a char for each link examined
        0 All zero's
        1 All one's
        2 Clock by 2
        4 Clock by 4
        L L1A
        E Event Header
        S Soft Reset
        B BC Reset
        J Junk (ie anything else)

     IPC:  Needs feedback, some delay to response
     Conf: Some lookup?!
  */
  std::vector<char> probe(unsigned int rod, 
                          signed int harness = -1);

  /// Send a trigger to all channels and see what comes back
  /**
     Do quick scan of channels
     Recognises all the same, clock by 2 and an event header
     Return a char for each link examined
        0 All zero's
        1 All one's
        2 Clock by 2
        4 Clock by 4
        L L1A
        E Event Header
        S Soft Reset
        B BC Reset
	{b,a,d, ,c,r,a,t,e} : Nothing can be returned -- no such crate! (only relevant over top level ipc)
        J Junk (ie anything else)

     IPC:  Needs feedback, some delay to response
     Conf: Some lookup?!
  */
  std::vector<char> probeWithTrigger(unsigned int rod, 
                                     const Trigger *trigger, signed int harness = -1);

  /// Scan over a variable and perform a probe for each
  std::vector<std::vector<char> > SctApi::probeScan(unsigned int rod, 
                                                    boost::shared_ptr<Scan> scan, signed int harness = -1);

  /**
     Probe all modules and check that value returned by probe 
     is one of the characters specified

     @param value Set of chars allowed as response to probe
   */
  bool checkAllModulesProbe(std::string value);

  /**
     Suggest values for configuration

     1) Perform probe scan of tx channels and interpret giving connections of modules to tx channels
   */
  void autoConfigure();

  /**
     Run the ROD "BOC histogram" primitive
   */
  void bocHistogram(unsigned int rod, 
                    unsigned int samples, unsigned int numLoops);


  /************ Direct DDC (no translations) ****************/
  SctApiDDC * getSctApiDDC() const;

  /**
     Stop all tasks to do with histogramming

     This is a lower level function than abort,
     which returns to a know state

     DDC:  If changed?
     IPC:  Any time
     Conf: No communication
   */
  void tidyHistogramming();

  /**
     Return slot this ROD is in
   */
  long getRodSlot(unsigned int rod);

  /**
     Return the revision of a ROD
   */
  int getRodRevision(unsigned int rod);

  /**
     Return the revision of a ROD
   */
  int getRodRevision(const RodLabel &label);

  /**
     Select link to send to ROD front panel LEMO
  */
  void testLinkOutSelect(unsigned int rod,
                         unsigned int link);

  /**
     Return selected link to send to ROD front panel LEMO
  */
  unsigned int testLinkOutRetrieve(unsigned int rod);

  /**
     Disable a debug option

     DDC:  No action 
     IPC:  Class action
     Conf: No communication
  */
  void unsetDebugOption(std::string opt);

  /**
     Enable a debug option

     DDC:  No action 
     IPC:  Class action
     Conf: No communication
  */
  void setDebugOption(std::string opt);

  /**
     Get the currently enabled debug options

     DDC:  No action 
     IPC:  Class action, immediate response
     Conf: No communication
  */
  std::list<std::string> listEnabledDebugOptions() const;

  /**
     List possible debug options

     DDC:  No action 
     IPC:  Class action, immediate response
     Conf: No communication
  */
  std::vector<std::string> listDebugOptions() const;

  /**
     For use in emergency situations only (not as a matter of course) use unsetBusy() to override SctApi's present busy state and ask it to appear "not busy"  to the outside world.  See also "SctApi::hasFinishedScanning()".
  */
#warning "Call this unset scan busy?"
  void unsetBusy();

  /**
     Perform a step in a histogram which is in a stalled state
   */
  void debugStepHistogram();

  /**
     Continue a histogram which was previously stalled

     This only works if the histogram is at the end of a bin
   */
  void debugContinueHistogram();

  /**
     Try to abort the current histogram

     This may not work. The DSP can be in a state where the abort signal is 
     delayed in its interpretation
   */
  void debugAbortHistogram();

  /**
     Setup event trapping

     A reduced set of parameters compared to the primitive is used.
   */
  void eventTrapSetup(int slaveMask, int trapMatch, int trapMod, int trapRemain, bool tim, bool error, 
                      boost::shared_ptr<PrimListWrapper> primList);

  /// Start event trap on a slave
  void startEventTrap(int slave, boost::shared_ptr<PrimListWrapper> primList);

  /// Stop event trap on a slave
  void stopEventTrap(int slave, boost::shared_ptr<PrimListWrapper> primList);

  /// Direct write of ROD mode!
  void rodMode(unsigned int rod,
               int mode, int flag, int fifoSetup, int nBins, int delay, int message);

  /// Set module masks
  void setupModuleMask(unsigned int rod,
                       int port, int slvs);

  /// Add set module masks command to primitive list
  void setupModuleMask(int port, int slvs, boost::shared_ptr<PrimListWrapper> list);

  /// Send a SEND_DATA primitive and retrieve the (pointer, length) result 
  std::pair<UINT32, UINT32> sendData(unsigned int rod, int type, int dsp);

  /**
     Resume the primitive polling thread
  */
  void resumePolling();

  /**
     Pause the primitive polling thread
  */
  void stopPolling();

  /// Print ROD calibration registers
  void print_calib(unsigned int rod);

  /// Send a primitive list to all the RODs in this crate
  /**
     A helper function for other methods in this class,
     to be used only internally to this class as
     it exposes the RodPrimList class
  */
  int synchSendPrimListAll(boost::shared_ptr<PrimListWrapper> primList, int timeout = 10);

  /// Stop the histogramming and event trapping tasks
  void stopHistogramming(const ScanEx &ex);

  /// Set block of slave memory to value (using RW_SLAVE_MEMORY)
  void setSlaveBlock(unsigned int rod, 
                     int s, long sAdd, int words, long value = 0, bool usePrim = true);

  /// Write to a ROD register
  void writeRODRegister(unsigned int rod,
                        int reg, int off, int width, int value);

  /// Read a register from the ROD
  unsigned int readRODRegister(unsigned int rod, int r);

  /// Poll a register
  int pollRegister(unsigned int rod, int r, int off, int width, int val, int timeout = 1000000);

  virtual unsigned long *readFifo(unsigned int rod, int id, int bank, int elems);

  /// Create histogram task primitive
  void startHistogramTask(const Scan &scan, const ScanEx &ex, const RodLabel, 
                          unsigned int startBin, unsigned int nBins, unsigned int nTrigs,
                          boost::shared_ptr<PrimListWrapper> list);

  /// Send a primitive list to a ROD
  /**
     A helper function for other methods in this class,
     to be used only internally to this class as 
     it exposes the RodPrimList class
  */
  void sendPrimList(unsigned int rod,
                    boost::shared_ptr<PrimListWrapper> prim);

  /// Send a primitive list to a ROD and wait for completion
  /**
     A helper function for other methods in this class,
     to be used only internally to this class as
     it exposes the RodPrimList class
  */
  int synchSendPrimList(unsigned int rod, boost::shared_ptr<PrimListWrapper> primList, int timeout = 10);

  MRSStream *mrs;

  /// needed by ScanControl to update the cache to the last point in the scan.
  SctApiConfigCache& getModuleConfigCache();
#if USE_IS
  boost::shared_ptr<ISInfoDictionary> m_isDict;
#endif

  unsigned preloadModuleConfigurations();
  volatile bool m_preloadModuleConfigurationsStop;
 protected:
  /** 
      version of setABCDModule which will not look at the crate controller's local cache 
      but will actually set it to the ROD.
   */
  virtual void really_setABCDModule(UINT32 mid, std::list<BankType> banks, UINT32 api_cfg_bits);
  
  /** 
      version of getABCDModule which will not look at the crate controller's local cache 
      but will actually get it from the ROD
   */
  virtual void really_getABCDModule(UINT32 mid, BankType bank);

  /********** Should these be part of the API? **************/
  

  ///Initialise a ROD from the configuration
  /**
      Proxy for Crate

      Return 0 for success
  */
  virtual int initialiseRod(unsigned int rod);

  /// Initialise a Module from the configuration.
  /**  
       Load module configuration from database
       Send module configuration to ROD (unless toRod is set to false)

       Return 0 for success
  */
  virtual int initialiseModule(std::string module, bool toRod=true);

  /// Lookup module configuration in the cache
  ABCDModule *lookupConfig(UINT32 mid);

  /// Scan methods
  /**
    Perform necessary hardware checks before a scan
    @return true if successful, false otherwise
    */
  virtual bool preScanHardwareCheck(const Scan& scan, ScanEx& extra);
  
  virtual void doHistogramSetup(const Scan &scan, const ScanEx &extra);
  
  /**
     IPC:  Needs feedback, immediate response
     Conf: No Communication
  */
  UINT32 *readEventBuffer(unsigned int rod, 
                        int sl, bool extFlag, bool errorType);

  /**
     Return the frame in which the last event starts
   */
  int getLastEventFrame(unsigned int rod, int slave);

  /// Poll the histogramming
  /**
     Wait for the number of bins to equal (bins - 1) on all master 
     dsps involved in the scan.

     Then wait for the number of events on the slave to equal to events

     Timeout if the bin count doesn't change, and if the number 
     of events doesn't change
  */
  virtual int pollHistogramming(boost::shared_ptr<ScanControl> controller, int timeout = 5);
 
  void scanLoop();

  void scanPollingThread();

  void addRodToList(RodLabel newRod);

  RodInfo &SctApi::getRodInfo(const RodLabel &l);

  /// List all the RODs
  std::list<RodLabel> rodList;

  /// List them again, but with additional info...
  std::map<RodLabel, RodInfo> rodInfoList;
  boost::recursive_mutex rodInfoListAccess;

  /// Cache of module configurations
  std::map<UINT32, ABCDModule> moduleMap;
 
  /// Controls access to module configurations during initialisation
  boost::mutex moduleMapAccess;
 
  // Controller for current scan
  boost::shared_ptr<ScanControl> scanController;

  // Possibly to be used for continuing a scan
  boost::shared_ptr<ScanControl> lastScanController;

 private:
  
  /*************************************************************
   *          Internal functions (shouldn't be exported,       *
   *              probably able to keep them private )         *
   *************************************************************/

  unsigned int prt() const; // shortcut to ucid().partition();
  unsigned int crt() const; // shortcut to ucid().crate();
  Sct::URID localRod(const Sct::RodIdentifier rod) const; // returns the URID of the supplied rod assuming that it lies within this crate.

  /// Lets raw scans run asynchronously
  /** Called by doRawScanThread
   */
  void doRawScanThreadInnards(boost::shared_ptr<Scan> scan, int delay, int width, 
                              bool configureModules, bool clkBy2);
  /// Lets raw scans run asynchronously
  /** Called by doRawScan
   */
  void doRawScanThread(boost::shared_ptr<Scan> scan, int delay, int width, 
                       bool configureModules, bool clkBy2);

  
  /// Send a primitive list to a slave DSP
  /** 
      Send a RodPrimList to the Master DSP which contains the RodPrimList which is passed as an argument

      @param await If true master waits for completion of slave primlist
      @param response If true master expects reponse from slave
   */
  void sendSlavePrimList(unsigned int rod,
                         boost::shared_ptr<PrimListWrapper> prim, unsigned int slave, bool await, bool response);

  /// Send a primitive list to all the RODs in a crate
  /**
     A helper function for other methods in this class,
     to be used only internally to this class as 
     it exposes the RodPrimList class
  */
  void sendPrimListAll(boost::shared_ptr<PrimListWrapper> prim);

  /// Setup event trapping according to scan and ex
  void setupEventTrapping(const Scan &scan, const ScanEx &ex, const RodLabel, boost::shared_ptr<PrimListWrapper> list);

  /// Start event trapping according to scan and ex
  void startEventTrapping(const Scan &scan, const ScanEx &ex, const RodLabel, boost::shared_ptr<PrimListWrapper> primList);

  /// Setup an event dump according to scan and ex
  void setupEventDump(const Scan &scan, const ScanEx &ex, const RodLabel, boost::shared_ptr<PrimListWrapper> list);

  /// Setup histogramming according to scan and ex
  void setupHistogramming(const Scan &scan, const ScanEx &ex, const RodLabel, bool tim, 
                          boost::shared_ptr<PrimListWrapper> list);

  /// Setup ROD for calibration
  void calib_init();

  /// Read raw data from all links into one "histogram"
  void readRawData(unsigned int rod, 
                   int delay, bool setMask);

  ///  Put raw data into FIFOs ready to read out
  virtual int setupRawData(unsigned int rod, 
			   int delay, int units, bool setMask, const Trigger *trig = 0);

  /// Utility function to get the Crate object from the map
  /** 
      Lookup the Crate object to send instructions to.
      Return 0 if not found.
   */
  Crate *getCrate() const;

  /// Put module mid into the cache (if present in the configuration)
  void cacheModuleConfig(UINT32 mid, std::string name="");

  /// Translate mid to pcrc of redundant connection
  /**
     XXX What does it do on failure!!!
   */
  void getrpcrc(UINT32 mid, unsigned int &rpartition, unsigned int &rcrate, unsigned int &rrod, unsigned &rchannel);

  /// Setup module, dsp masks etc
  void setupScanMasks(ScanEx &extra, int distSlave, bool dual);

  /// Return whether check passed
  bool checkModuleListsForScan();

  /// Setup modules before scanning
  /**
     Needs scan to know whether to set formatters into expanded mode
   */
  void preScanModuleSetup(const Scan &scan);

  /**
     Insert SctApiCounters into IS
   */
  void insertISStats();

  /**
     Update the IS version of SctApiCounters
   */
  void updateISStats();

  // AJB needed to make this public for ScanControl to use
 public:
  bool checkDebugOption(std::string opt) const;
  // This should be DebugOptions but that would mean including SctApiDebug.h here
  bool checkDebugOption(int opt) const;
 private:

  /*********** Internal variables *************/

  /// This Idiosyncrasy is what makes us different from every other crate controller
  boost::shared_ptr<Idiosyncrasy> m_id; // NB this is shared to hide the implementaton, not because we share the ownership with anyone.  The same is true for most of the private members below:

  /// The configuration database
  boost::shared_ptr<SctConfiguration::Configuration> config;

  boost::shared_ptr<DCSAccess> m_dcsAccess;
  boost::shared_ptr<SctApiDDC> m_sctApiDDC;

  /// The current debug primitive list
  boost::shared_ptr<PrimListWrapper> debugPrimList;

  /// Map to retrieve crate objects
  //std::map<std::pair<unsigned int, unsigned int>, Crate* > crateMap;
  boost::shared_ptr<Crate> crateObject;

  /// Log everything that is called
  boost::shared_ptr<Log> m_log;
  Log & log() { return *m_log; };

  /// For debug only, save the scan information
  boost::shared_ptr<const Scan> lastDebugScan;
  boost::shared_ptr<const ScanEx> lastDebugScanEx;

  /// Current scan number
  UINT32 scanNumber;

  /// Current run number
  UINT32 runNumber;

  /// Current run mode
  RunType currentRunMode;
  
  /// list of banks to automatically update on ROD
  std::list<BankType> m_autoUpdateBanks;

#if USE_SCAN_THREAD
  boost::condition scanQueue_notEmpty;
  boost::mutex scanQueue_mutex;

  /* ******* Threads ******/
  std::auto_ptr<boost::thread> scanPollThread;

  bool m_stopPolling;
  bool m_inScanLoop;
  bool m_inRawScanLoop;
#endif

  /** configuration cache **/
  SctApiConfigCache module_cache;

  /* **********  Mutexes  ********** */
};
}  // End of namespace SctApi
#endif
