#ifndef SCT_API_CONFIG_CACHE_H
#define SCT_API_CONFIG_CACHE_H
/** 
    Cache for recording what the Api thinks lives in the ROD 
    configuration banks for module configurations.
    It can use this information to reduce the information it needs
    to transfer over the VME back plane to the RODS.
*/
#include <Sct/AbcdModule.h>
#include <CommonWithDsp/processor.h>
#include <boost/shared_ptr.hpp>
#include <string>
#include <map>
#include "SctApiFwd.h"
#include <boost/thread.hpp>

namespace SctApi {
/**
   This is supposed to mirror the module configuration banks on the rod.
   It allows the API to check what the configuration looks like,
   and to only do updates of those parts which are different.
*/
class SctApiConfigCache{
 public:
  class ModuleBanks;
  /**
     return shared pointer to the ABCDModule structure,
     or null shared pointer if it dosent exist.
  */
  boost::shared_ptr<ModuleBanks> getFromMid(const UINT32 mid);
  boost::shared_ptr<const ModuleBanks> getFromMid(const UINT32 mid) const;
  /**
     clear a module from the cache
   */
  void clearMid(const UINT32 mid);
  /**
     clear all modules.
   */
  void clearAll();

  /// copy from single bank to a set of banks for all modules
  void copyAllModules(BankType from, std::list<BankType> to);
  /// copy from single bank to single bank for all modules
  void copyAllModules(BankType from, BankType to);
  /** 
      add a module with a particular MID and bank
  */
  void set(UINT32 mid, boost::shared_ptr<ModuleBanks> banks);
  /**
     constructor.
   */
  SctApiConfigCache();
   /**
     destructor
   */
  ~SctApiConfigCache();
  /**
     Class to hold different ABCDModule configurations for a particular module.
   */
  class ModuleBanks{
  public:
    /** 
	get a config for a particular BankType 
	@return null pointer if not in map 
    */
    boost::shared_ptr<ABCDModule> get(BankType b);
    /** get a config for a particular BankType (const) 
	@return null pointer if not in map
     */
    boost::shared_ptr<const ABCDModule> get(BankType b) const;
    /** 
	make deep copy of a particular bank , using cloneABCDModule
	if from bank does not exist, then nothing happens
    */
    void copy(BankType from, BankType to);
    /** 
	make deep copy of a particular bank , using cloneABCDModule
	if from bank does not exist, then nothing happens
    */
    void copy(BankType from, std::list<BankType> to);
    /** set contents of a bank */
    void set(BankType, boost::shared_ptr<ABCDModule> config);
  private:
    std::map<BankType, boost::shared_ptr<ABCDModule> > m_map;
    mutable boost::recursive_mutex bank_mutex;
  };

 private:
  std::map<UINT32, boost::shared_ptr<ModuleBanks> > m_map;  ///< The map itself
  mutable boost::recursive_mutex cache_mutex;
};

}

#endif
