
#ifndef SCT_SCTAPI_TRIGGER_H
#define SCT_SCTAPI_TRIGGER_H

//includes
#include <utility> // for std::pair
#include <vector>

#include "CommonWithDsp/processor.h"

#include "config.h"

//defn
namespace SctApi {

  class Trigger {
  public:
    virtual ~Trigger();

    /// Type for representation as a series of pairs as defined for ROD DSP
    typedef std::pair<UINT16, UINT32>  RODTrigger;
    typedef std::vector<RODTrigger>    RODTriggers;
    typedef UINT16 TIMTrigger;
    typedef std::vector<UINT16> TIMTriggers;

    enum Source{SOURCE_UNKNOWN, SOURCE_ROD,
		SOURCE_TIM_INTERNAL, SOURCE_TIM_EXTERNAL};
    
    /// Set to a single L1A
    virtual void singleL1A() = 0;
    /// Set to a double L1A
    virtual void doubleL1A(unsigned short delay) = 0;
    /// Set to a delayed L1A
    virtual void delayedL1A(unsigned short delay) = 0;
    /// Set to a Calibration pulse followed by an L1A BC0s later
    virtual void calL1A(unsigned short delay) = 0;
    virtual void pulseL1A(unsigned short delay) = 0;
    virtual void softL1A(unsigned short delay) = 0;
    virtual void softCalL1A(unsigned short delay, unsigned short delay2) = 0;
    virtual void softPulseL1A(unsigned short delay, unsigned short delay2) = 0;
    virtual void softBc(unsigned short delay) = 0;
    virtual void soft() = 0;
    virtual void bc() = 0;
    virtual void bcL1A(unsigned short delay) = 0;
    virtual void bcCalL1A(unsigned short delay, unsigned short delay2) = 0;
    virtual void bcPulseL1A(unsigned short delay, unsigned short delay2) = 0;

    virtual void setCommIncr(unsigned short command, unsigned short incr) = 0;
    virtual void getCommIncr(unsigned short &command, unsigned short &incr) const = 0;

    virtual void clear() = 0;
    virtual void addCommand(UINT16 first, UINT32 second) = 0;

    virtual Source getSource() const = 0;
    virtual void setSource(Source) = 0;
    
    /// kHz
    virtual void setFrequency(double freq) = 0;
    /// kHz
    virtual double getFrequency() const = 0;

    // what sort of randomiser to use?
    virtual int getRandom() const = 0;
    virtual void setRandom(int) = 0;

    /// Get list of triggers in ROD format
    virtual const RODTriggers getRODTriggers(int scan_point=0) const = 0; 
    virtual const TIMTriggers getTIMTriggers(int scan_point=0) const = 0;

    virtual std::string print() const = 0;

    virtual bool isValidRODTrigger() const = 0;

    virtual bool isValidTIMTrigger() const = 0;

    virtual bool timNeedsSequencer() const = 0;

  /**
     Build a TIM trigger sequence from a ROD trigger.
     @note should only do this if it is a valid TIMTrigger.
     @throw SctApiException if it is not a valid TIM trigger.
  */
  static ::SctApi::Trigger::TIMTriggers convertToTIMTriggers(const ::SctApi::Trigger::RODTriggers& rTrigs);

  /**
     TIM can only send a single "CAL" line, which could mean any of the ABCD slow commands.
     This function finds out which of the ABCD slow commands the TIM-CAL corresponds to
     @return slow command ROD trigger type, or zero if no slow command is requested.
     @throw SctApiException if more than one type of slow command is requested.
  */
  static int getUniqueSlowCommandType(const ::SctApi::Trigger::RODTriggers& rTrigs);

  /**
     
   */
  static std::string getNameOfRODTrigger(RODTrigger t);
};

}
#endif
